import csv
import re
from collections.abc import Mapping
from copy import deepcopy
from datetime import date, datetime
from pathlib import Path
from typing import ClassVar

import pyarrow.parquet as pq
import pytest
import yaml
from cognite.client.data_classes.data_modeling import ContainerId, PropertyId, ViewId
from cognite.client.data_classes.data_modeling import data_types as dt
from cognite.client.data_classes.data_modeling.data_types import DirectRelationReference, EnumValue
from cognite.client.data_classes.data_modeling.views import (
    MappedProperty,
    MultiEdgeConnection,
    MultiReverseDirectRelation,
    SingleEdgeConnection,
    SingleReverseDirectRelation,
    ViewProperty,
)

from cognite_toolkit._cdf_tk.utils.table_writers import (
    CSVWriter,
    ParquetWriter,
    Rows,
    Schema,
    SchemaColumn,
    SchemaColumnList,
    YAMLWriter,
)


@pytest.fixture()
def example_schema() -> Schema:
    return Schema(
        display_name="Example",
        folder_name="example",
        kind="Example",
        format_="csv",
        columns=SchemaColumnList(
            [
                SchemaColumn(name="myString", type="string"),
                SchemaColumn(name="myInteger", type="integer"),
                SchemaColumn(name="myFloat", type="float"),
                SchemaColumn(name="myBoolean", type="boolean"),
                SchemaColumn(name="myJson", type="json"),
                SchemaColumn(name="myTimestamp", type="timestamp"),
                SchemaColumn(name="myDate", type="date"),
                SchemaColumn(name="myStringList", type="string", is_array=True),
                SchemaColumn(name="myIntegerList", type="integer", is_array=True),
                SchemaColumn(name="myFloatList", type="float", is_array=True),
                SchemaColumn(name="myBooleanList", type="boolean", is_array=True),
                SchemaColumn(name="myJsonList", type="json", is_array=False),
                SchemaColumn(name="myTimestampList", type="timestamp", is_array=True),
                SchemaColumn(name="myDateList", type="date", is_array=True),
            ]
        ),
    )


@pytest.fixture()
def example_data() -> Rows:
    return [
        {
            "myString": "value1",
            "myInteger": 1,
            "myFloat": 1.0,
            "myBoolean": True,
            "myJson": {"key": "value"},
            "myTimestamp": "2023-01-01T00:00:00Z",
            "myDate": "2023-01-01",
            "myStringList": ["on", "off", '"maybe"', "y", "n"],
            "myIntegerList": [1, 2],
            "myFloatList": [1.0, 2.0],
            "myBooleanList": [True, False],
            "myJsonList": [{"key1": "value1"}, {"key2": "value2"}],
            "myTimestampList": ["2023-01-01T00:00:00Z", "2023-01-02T00:00:00Z"],
            "myDateList": ["2023-01-01", "2023-01-02"],
        },
        {
            "myString": "value2",
            "myInteger": None,
            "myFloat": 2.0,
            "myBoolean": False,
            "myJson": None,
            "myTimestamp": None,
            "myDate": None,
            "myStringList": None,
            "myIntegerList": None,
            "myFloatList": None,
            "myBooleanList": None,
            "myJsonList": None,
            "myTimestampList": None,
            "myDateList": None,
        },
        {
            "myString": None,
            "myInteger": 3,
            "myFloat": None,
            "myBoolean": True,
            "myJson": {"key": "value"},
            "myTimestamp": datetime(2023, 1, 3, 0, 0, 0),
            "myDate": date(2023, 1, 3),
            "myStringList": ["yes", "no"],
            "myIntegerList": [3, 4],
            "myFloatList": [3.0, 4.0],
            "myBooleanList": [True, False],
            "myJsonList": [{"key1": "value1"}, {"key2": "value2"}],
            "myTimestampList": [datetime(2023, 1, 3, 0, 0, 0), datetime(2023, 1, 4, 0, 0, 0)],
            "myDateList": [date(2023, 1, 3), date(2023, 1, 4)],
        },
    ]


class TestTableFileWriter:
    part_pattern = r"part-(?P<part>\d{4})"

    def get_part_number(self, file_path: Path) -> int:
        """Extract the part number from the file name."""
        # This is needed to sort files as Windows and Linux sorts numbers in file names differently.
        match = re.search(self.part_pattern, file_path.name)
        if match:
            return int(match.group("part"))
        return 0

    def test_write_csv(self, example_schema: Schema, example_data: Rows, tmp_path: Path) -> None:
        output_dir = tmp_path / "output"

        with CSVWriter(example_schema, output_dir) as writer:
            for single_row in example_data:
                writer.write_rows([("group1", [single_row])])

        csv_file = list(output_dir.rglob("*.csv"))
        assert len(csv_file) == 1
        actual_data: Rows = []
        with csv_file[0].open("r", newline="") as f:
            reader = csv.reader(f)
            header = next(reader)
            for row in reader:
                actual_data.append(dict(zip(header, row)))

        # csv writer stringifies all values, so we convert them back to their original types.
        # Note at this stage (2. June 2025) Toolkit does not support a CSV reader.
        expected = [
            {key: str(value if value is not None else "") for key, value in row.items()} for row in example_data
        ]
        assert actual_data == expected

    def test_write_yaml(self, example_schema: Schema, example_data: Rows, tmp_path: Path) -> None:
        output_dir = tmp_path / "output"
        example_schema.format_ = "yaml"

        with YAMLWriter(example_schema, output_dir) as writer:
            for single_row in example_data:
                writer.write_rows([("group1", [single_row])])

        yaml_file = list(output_dir.rglob("*.yaml"))
        assert len(yaml_file) == 1
        actual_data: Rows = yaml.safe_load(yaml_file[0].read_text(encoding="utf-8"))
        assert actual_data == example_data

    def test_write_parquet(self, example_schema: Schema, example_data: Rows, tmp_path: Path) -> None:
        output_dir = tmp_path / "output"
        example_schema.format_ = "parquet"
        # The parquet writer mutates the input data, so we need to deepcopy it.
        example_data_copy = deepcopy(example_data)

        with ParquetWriter(example_schema, output_dir) as writer:
            for single_row in example_data_copy:
                writer.write_rows([("group1", [single_row])])

        parquet_file = list(output_dir.rglob("*.parquet"))
        assert len(parquet_file) == 1
        table = pq.read_table(parquet_file[0])
        assert table.num_rows == len(example_data)
        actual_data = table.to_pylist()
        # We compare the mutated data, as the ParquetWriter standardizes the date, timestamp, and json formats.
        assert actual_data == example_data_copy

    def test_write_csv_above_limit(self, example_schema: Schema, example_data: Rows, tmp_path: Path) -> None:
        output_dir = tmp_path / "output"
        with CSVWriter(example_schema, output_dir, max_file_size_bytes=1) as writer:
            for single_row in example_data:
                writer.write_rows([("group1", [single_row])])

        csv_files = sorted(output_dir.rglob("*.csv"), key=self.get_part_number)
        assert len(csv_files) == len(example_data)
        # Each row should be written to a separate file due to the size limit
        for csv_file, row in zip(csv_files, example_data):
            with csv_file.open("r", newline="") as f:
                reader = csv.reader(f)
                header = next(reader)
                actual_row = dict(zip(header, next(reader)))
                # csv writer stringifies all values, so we convert them back to their original types.
                expected_row = {key: str(value if value is not None else "") for key, value in row.items()}
                assert actual_row == expected_row

    def test_write_parquet_above_limit(self, example_schema: Schema, example_data: Rows, tmp_path: Path) -> None:
        output_dir = tmp_path / "output"
        # The parquet writer mutates the input data, so we need to deepcopy it.
        example_data_copy = deepcopy(example_data)
        with ParquetWriter(example_schema, output_dir, max_file_size_bytes=1) as writer:
            for single_row in example_data_copy:
                writer.write_rows([("group1", [single_row])])

        parquet_files = sorted(output_dir.rglob("*.parquet"), key=self.get_part_number)
        assert len(parquet_files) == len(example_data)
        # Each row should be written to a separate file due to the size limit
        # We compare the mutated data, as the ParquetWriter standardizes the date, timestamp, and json formats.
        for parquet_file, row in zip(parquet_files, example_data_copy):
            table = pq.read_table(parquet_file)
            assert table.num_rows == 1
            assert table.to_pylist() == [row]

    def test_write_yaml_above_limit(self, example_schema: Schema, example_data: Rows, tmp_path: Path) -> None:
        output_dir = tmp_path / "output"
        with YAMLWriter(example_schema, output_dir, max_file_size_bytes=1) as writer:
            for single_row in example_data:
                writer.write_rows([("group1", [single_row])])

        yaml_files = sorted(output_dir.rglob("*.yaml"), key=self.get_part_number)
        assert len(yaml_files) == len(example_data)
        for yaml_file, row in zip(yaml_files, example_data):
            with yaml_file.open("r", encoding="utf-8") as f:
                actual_data = yaml.safe_load(f)
                assert actual_data == [row]


class TestSchemaColumnList:
    container: ClassVar[ContainerId] = ContainerId("my_space", "my_container")
    view: ClassVar[ViewId] = ViewId("my_space", "my_view", "v1")
    default_values: ClassVar[Mapping[str, bool]] = dict(nullable=True, immutable=False, auto_increment=False)
    all_view_properties: Mapping[str, ViewProperty] = {
        "myText": MappedProperty(container, "myText", dt.Text(), **default_values),
        "myBoolean": MappedProperty(container, "myBoolean", dt.Boolean(), **default_values),
        "myTimestamp": MappedProperty(container, "myTimestamp", dt.Timestamp(), **default_values),
        "myDate": MappedProperty(container, "myDate", dt.Date(), **default_values),
        "myFloat32": MappedProperty(container, "myFloat32", dt.Float32(), **default_values),
        "myFloat64": MappedProperty(container, "myFloat64", dt.Float64(), **default_values),
        "myInt32": MappedProperty(container, "myInt32", dt.Int32(), **default_values),
        "myInt64": MappedProperty(container, "myInt64", dt.Int64(), **default_values),
        "myJson": MappedProperty(container, "myJson", dt.Json(), **default_values),
        "myStringList": MappedProperty(container, "myStringList", dt.Text(is_list=True), **default_values),
        "myIntegerList": MappedProperty(container, "myIntegerList", dt.Int64(is_list=True), **default_values),
        "myFloatList": MappedProperty(container, "myFloatList", dt.Float64(is_list=True), **default_values),
        "myBooleanList": MappedProperty(container, "myBooleanList", dt.Boolean(is_list=True), **default_values),
        "myJsonList": MappedProperty(container, "myJsonList", dt.Json(is_list=False), **default_values),
        "myDirectRelation": MappedProperty(container, "myDirectRelation", dt.DirectRelation(), **default_values),
        "myDirectRelationList": MappedProperty(
            container, "myDirectRelationList", dt.DirectRelation(is_list=True), **default_values
        ),
        "myTimeSeriesReference": MappedProperty(
            container, "myTimeSeriesReference", dt.TimeSeriesReference(), **default_values
        ),
        "myFileReference": MappedProperty(container, "myFileReference", dt.FileReference(), **default_values),
        "mySequenceReference": MappedProperty(
            container, "mySequenceReference", dt.SequenceReference(), **default_values
        ),
        "myTimeSeriesReferenceList": MappedProperty(
            container, "myTimeSeriesReferenceList", dt.TimeSeriesReference(is_list=True), **default_values
        ),
        "myFileReferenceList": MappedProperty(
            container, "myFileReferenceList", dt.FileReference(is_list=True), **default_values
        ),
        "mySequenceReferenceList": MappedProperty(
            container, "mySequenceReferenceList", dt.SequenceReference(is_list=True), **default_values
        ),
        "myEnum": MappedProperty(
            container, "myEnum", dt.Enum({"value1": EnumValue(), "value2": EnumValue()}), **default_values
        ),
        "myMultiEdge": MultiEdgeConnection(
            DirectRelationReference("my_space", "myMultiEdge"), view, None, None, None, "outwards"
        ),
        "mySingleEdge": SingleEdgeConnection(
            DirectRelationReference("my_space", "mySingleEdge"), view, None, None, None, "outwards"
        ),
        "myMultiReverseDirectRelation": MultiReverseDirectRelation(
            view, PropertyId(view, "myMultiReverseDirectRelation"), None, None
        ),
        "mySingleReverseDirectRelation": SingleReverseDirectRelation(
            view, PropertyId(view, "mySingleReverseDirectRelation"), None, None
        ),
    }

    def test_create_schema_from_view_properties(self) -> None:
        columns = SchemaColumnList.create_from_view_properties(self.all_view_properties)
        assert columns == SchemaColumnList(
            [
                SchemaColumn("space", "string"),
                SchemaColumn("externalId", "string"),
                SchemaColumn("instanceType", "string"),
                SchemaColumn("existingVersion", "integer"),
                SchemaColumn("type", "json"),
                SchemaColumn(name="properties.myText", type="string"),
                SchemaColumn(name="properties.myBoolean", type="boolean"),
                SchemaColumn(name="properties.myTimestamp", type="timestamp"),
                SchemaColumn(name="properties.myDate", type="date"),
                SchemaColumn(name="properties.myFloat32", type="float"),
                SchemaColumn(name="properties.myFloat64", type="float"),
                SchemaColumn(name="properties.myInt32", type="integer"),
                SchemaColumn(name="properties.myInt64", type="integer"),
                SchemaColumn(name="properties.myJson", type="json"),
                SchemaColumn(name="properties.myStringList", type="string", is_array=True),
                SchemaColumn(name="properties.myIntegerList", type="integer", is_array=True),
                SchemaColumn(name="properties.myFloatList", type="float", is_array=True),
                SchemaColumn(name="properties.myBooleanList", type="boolean", is_array=True),
                SchemaColumn(name="properties.myJsonList", type="json", is_array=False),
                SchemaColumn(name="properties.myDirectRelation", type="json"),
                SchemaColumn(name="properties.myDirectRelationList", type="json", is_array=False),
                SchemaColumn(name="properties.myTimeSeriesReference", type="string"),
                SchemaColumn(name="properties.myFileReference", type="string"),
                SchemaColumn(name="properties.mySequenceReference", type="string"),
                SchemaColumn(name="properties.myTimeSeriesReferenceList", type="string", is_array=True),
                SchemaColumn(name="properties.myFileReferenceList", type="string", is_array=True),
                SchemaColumn(name="properties.mySequenceReferenceList", type="string", is_array=True),
                SchemaColumn(name="properties.myEnum", type="string"),
            ]
        )

    def test_create_schema_from_view_properties_with_edge_support(self) -> None:
        columns = SchemaColumnList.create_from_view_properties(self.all_view_properties, support_edges=True)

        assert {SchemaColumn("startNode", "json"), SchemaColumn("endNode", "json")} <= set(columns)



import itertools
import json
import time
import traceback
from collections.abc import Iterable, MutableSequence, Sequence
from datetime import datetime, timezone
from hashlib import sha256
from typing import Any, Literal, cast

from cognite.client.config import global_config

# Do not warn the user about feature previews from the Cognite-SDK we use in Toolkit
# ruff: noqa: E402
global_config.disable_pypi_version_check = True
global_config.silence_feature_preview_warnings = True
import yaml
from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes import ExtractionPipelineRunWrite, RowWrite
from cognite.client.data_classes.contextualization import ContextualizationJob
from cognite.client.data_classes.data_modeling.cdm.v1 import CogniteAnnotationApply
from pydantic import BaseModel, Field, field_validator, model_validator
from pydantic.alias_generators import to_camel

FUNCTION_ID = "contextualization_entity_matcher"
EXTRACTION_PIPELINE_EXTERNAL_ID = "ctx_entity_matching"
EXTERNAL_ID_LIMIT = 256
EXTRACTION_RUN_MESSAGE_LIMIT = 1000
EDGE_TYPE = "entity.match"


def handle(data: dict, client: CogniteClient) -> dict:
    try:
        execute(data, client)
    except Exception as e:
        tb = traceback.extract_tb(e.__traceback__)
        last_entry_this_file = next((entry for entry in reversed(tb) if entry.filename == __file__), None)
        suffix = ""
        if last_entry_this_file:
            suffix = f" in function {last_entry_this_file.name} on line {last_entry_this_file.lineno}: {last_entry_this_file.line}"

        status: Literal["failure", "success"] = "failure"
        # Truncate the error message to 1000 characters the maximum allowed by the API
        prefix = f"ERROR {FUNCTION_ID}: "
        error_msg = f'"{e!s}"'
        message = prefix + error_msg + suffix
        if len(message) >= EXTRACTION_RUN_MESSAGE_LIMIT:
            error_msg = error_msg[: EXTRACTION_RUN_MESSAGE_LIMIT - len(prefix) - len(suffix) - 3 - 1]
            message = prefix + error_msg + '..."' + suffix
    else:
        status = "success"
        message = FUNCTION_ID

    client.extraction_pipelines.runs.create(
        ExtractionPipelineRunWrite(extpipe_external_id=EXTRACTION_PIPELINE_EXTERNAL_ID, status=status, message=message)
    )
    return {"status": status, "message": message}


################# Data Classes #################


# Configuration classes
class Parameters(BaseModel, alias_generator=to_camel):
    auto_approval_threshold: float = Field(gt=0.0, le=1.0)
    auto_reject_threshold: float = Field(gt=0.0, le=1.0)
    feature_type: Literal[
        "simple", "insensitive", "bigram", "frequencyweightedbigram", "bigramextratokenizers", "bigramcombo"
    ]


class ViewProperties(BaseModel, alias_generator=to_camel):
    space: str
    external_id: str
    version: str
    properties: list[str]

    def as_view_id(self) -> dm.ViewId:
        return dm.ViewId(space=self.space, external_id=self.external_id, version=self.version)


class MatchingJobConfig(BaseModel, alias_generator=to_camel):
    source_view: ViewProperties
    target_views: list[ViewProperties]


class ConfigData(BaseModel, alias_generator=to_camel):
    instance_spaces: list[str]
    annotation_space: str
    matching_jobs: dict[str, MatchingJobConfig]


class ConfigState(BaseModel, alias_generator=to_camel):
    raw_database: str
    raw_table: str


class Config(BaseModel, alias_generator=to_camel):
    parameters: Parameters
    data: ConfigData
    source_system: dm.DirectRelationReference
    state: ConfigState

    @classmethod
    @field_validator("source_system", mode="before")
    def pares_direct_relation(cls, value: Any) -> Any:
        if isinstance(value, dict):
            return dm.DirectRelationReference.load(value)
        return value


class Cursors:
    def __init__(self, client: CogniteClient, state: ConfigState) -> None:
        self._cursor_by_key: dict[str, str | None] = {}
        self._client = client
        self._raw_database = state.raw_database
        self._raw_table = state.raw_table

    def get_cursor(self, key: str) -> str | None:
        if key not in self._cursor_by_key:
            self._cursor_by_key[key] = self._lookup_cursor(key)
        return self._cursor_by_key[key]

    def set_cursor(self, key: str, cursor: str | None) -> None:
        self._cursor_by_key[key] = cursor

    def _lookup_cursor(self, key: str) -> str | None:
        row = self._client.raw.rows.retrieve(db_name=self._raw_database, table_name=self._raw_table, key=key)
        if row is None or row.columns is None:
            return None
        return row.columns.get("cursor")

    def store(self, key: str) -> None:
        cursor = self.get_cursor(key)
        self._client.raw.rows.insert(
            db_name=self._raw_database,
            table_name=self._raw_table,
            row=RowWrite(key=key, columns={"cursor": cursor}),
        )


class Entity(BaseModel, alias_generator=to_camel, extra="allow", populate_by_name=True):
    node_id: dm.NodeId
    view: dm.ViewId
    standardized_properties: dict[str, str]
    name_by_alias: dict[str, str]

    @model_validator(mode="before")
    def pack_properties(cls, values: dict[str, Any]) -> dict[str, Any]:
        if "standardized_properties" in values or "standardizedProperties" in values:
            return values
        standardized_properties: dict[str, str] = {}
        for key in list(values.keys()):
            if key.startswith("prop"):
                standardized_properties[key] = values.pop(key)
        return {**values, "standardized_properties": standardized_properties}

    @field_validator("node_id", "view", "name_by_alias", mode="before")
    def load_json(cls, value: Any) -> Any:
        if isinstance(value, str):
            return json.loads(value)
        return value

    @field_validator("node_id", mode="before")
    def load_node_id(cls, value: Any) -> Any:
        if isinstance(value, dict):
            return dm.NodeId.load(value)
        return value

    @field_validator("view", mode="before")
    def load_view_id(cls, value: Any) -> Any:
        if isinstance(value, dict):
            return dm.ViewId.load(value)
        return value

    @classmethod
    def from_node(cls, node: dm.Node, properties: list[str]) -> "Entity":
        if not node.properties:
            raise ValueError(f"Node {node.as_id()} does not have properties")
        view_id, node_properties = next(iter(node.properties.items()))
        standardized_properties: dict[str, str] = {}
        name_by_alias: dict[str, str] = {}
        for no, prop in enumerate(properties):
            if prop in node_properties:
                # We standardize the property names to prop0, prop1, prop2, ...
                # This is such that we can easily match the properties to multiple target entities
                alias = f"prop{no}"
                standardized_properties[alias] = str(node_properties[prop])
                name_by_alias[alias] = prop
        return cls(
            node_id=node.as_id(),
            view=view_id,
            standardized_properties=standardized_properties,
            name_by_alias=name_by_alias,
        )

    @classmethod
    def from_annotation(cls, data: dict[str, Any]) -> "list[Entity]":
        return [cls.model_validate(item) for item in data["entities"]]

    def dump(self) -> dict[str, Any]:
        return {
            "nodeId": json.dumps(self.node_id.dump()),
            "view": json.dumps(self.view.dump()),
            **self.standardized_properties,
            "nameByAlias": json.dumps(self.name_by_alias),
        }


class EntityList(list, MutableSequence[Entity]):
    @property
    def unique_properties(self) -> set[str]:
        return set().union(*[entity.standardized_properties.keys() for entity in self])

    @classmethod
    def from_nodes(cls, nodes: Sequence[dm.Node], properties: list[str]) -> "EntityList":
        return cls([Entity.from_node(node, properties) for node in nodes])

    def dump(self) -> list[dict[str, Any]]:
        return [entity.dump() for entity in self]

    def property_product(self, other: "EntityList") -> list[tuple[str, str]]:
        return [
            (source, target) for source, target in itertools.product(self.unique_properties, other.unique_properties)
        ]


class MatchItem(BaseModel, alias_generator=to_camel):
    score: float
    target: Entity


class MatchResult(BaseModel, alias_generator=to_camel):
    source: Entity
    matches: list[MatchItem]

    @property
    def best_match(self) -> MatchItem | None:
        return max(self.matches, key=lambda match: match.score) if self.matches else None

    @classmethod
    def load(cls, data: dict[str, Any]) -> "MatchResult":
        return cls.model_validate(data)


# Logger using print
class CogniteFunctionLogger:
    def __init__(self, log_level: Literal["DEBUG", "INFO", "WARNING", "ERROR"] = "INFO"):
        self.log_level = log_level.upper()

    def _print(self, prefix: str, message: str) -> None:
        if "\n" not in message:
            print(f"{prefix} {message}")
            return
        lines = message.split("\n")
        print(f"{prefix} {lines[0]}")
        prefix_len = len(prefix)
        for line in lines[1:]:
            print(f"{' ' * prefix_len} {line}")

    def debug(self, message: str) -> None:
        if self.log_level == "DEBUG":
            self._print("[DEBUG]", message)

    def info(self, message: str) -> None:
        if self.log_level in ("DEBUG", "INFO"):
            self._print("[INFO]", message)

    def warning(self, message: str) -> None:
        if self.log_level in ("DEBUG", "INFO", "WARNING"):
            self._print("[WARNING]", message)

    def error(self, message: str) -> None:
        self._print("[ERROR]", message)


#####################################################

################# Functions #################


def execute(data: dict, client: CogniteClient) -> None:
    logger = CogniteFunctionLogger(data.get("logLevel", "INFO"))  # type: ignore[arg-type]
    logger.debug("Starting entity matching")
    config = load_config(client, logger)
    logger.debug("Loaded config successfully")

    cursors = Cursors(client, config.state)
    job_by_name = trigger_matching_jobs(client, cursors, config, logger)
    job_name_by_id = {job.job_id: name for name, job in job_by_name.items()}
    jobs = list(job_by_name.values())
    logger.info(f"Matching jobs triggered: {len(job_by_name)}")

    annotation_count = 0
    for completed_job in wait_for_completion(jobs, logger):
        if completed_job.error_message:
            logger.error(f"Job {completed_job.job_id} entity matching failed: \n  - {completed_job.error_message}")
            continue
        annotation_list = create_annotations(
            completed_job, config.data.annotation_space, config.source_system, config.parameters, logger
        )
        write_annotations(annotation_list, client, completed_job.job_id or 0, logger)
        annotation_count += len(annotation_list)
        job_name = job_name_by_id[cast(int, completed_job.job_id)]
        cursors.store(job_name)

    logger.info(f"Annotations created: {annotation_count}")


def trigger_matching_jobs(
    client: CogniteClient, cursors: Cursors, config: Config, logger: CogniteFunctionLogger
) -> dict[str, ContextualizationJob]:
    instance_spaces = config.data.instance_spaces
    jobs: dict[str, ContextualizationJob] = {}

    for job_name, job_config in config.data.matching_jobs.items():
        target_entities = EntityList()
        for target_view in job_config.target_views:
            target_nodes = client.data_modeling.instances.list(
                instance_type="node",
                sources=[target_view.as_view_id()],
                space=instance_spaces,
                limit=-1,
            )
            target_entities.extend(EntityList.from_nodes(target_nodes, target_view.properties))

        last_cursor = cursors.get_cursor(job_name)

        query = _create_query(job_config.source_view, instance_spaces, last_cursor, job_name)
        query_result = client.data_modeling.instances.sync(query)
        cursors.set_cursor(job_name, query_result.cursors[job_name])
        source_nodes = query_result.get_nodes(job_name)

        source_entities = EntityList.from_nodes(source_nodes, job_config.source_view.properties)
        combinations = source_entities.property_product(target_entities)
        unsupervised_model = client.entity_matching.fit(
            sources=source_entities.dump(),
            targets=target_entities.dump(),
            feature_type=config.parameters.feature_type,
            match_fields=combinations,
        )
        job = unsupervised_model.predict(
            sources=source_entities.dump(),
            targets=target_entities.dump(),
            num_matches=1,
            score_threshold=config.parameters.auto_reject_threshold,
        )
        jobs[job_name] = job
        logger.debug(f"Triggered matching job {job_name} with {len(source_entities)} entities")
    return jobs


def _create_query(
    view: ViewProperties, instance_spaces: list[str], last_cursor: str | None, name: str
) -> dm.query.Query:
    view_id = view.as_view_id()
    is_selected = dm.filters.And(
        dm.filters.In(["node", "space"], instance_spaces),
        dm.filters.HasData(views=[view_id]),
    )
    return dm.query.Query(
        with_={
            name: dm.query.NodeResultSetExpression(
                filter=is_selected,
                limit=1000,
            )
        },
        select={name: dm.query.Select([dm.query.SourceSelector(source=view_id, properties=view.properties)])},
        cursors={name: last_cursor},
    )


def wait_for_completion(
    jobs: list[ContextualizationJob], logger: CogniteFunctionLogger
) -> Iterable[ContextualizationJob]:
    # The Cognite Function will eventually time out, so we don't need to worry about running forever
    while jobs:
        job = jobs.pop(0)

        job.update_status()

        status = cast(str, job.status).casefold()
        if status == "completed":
            yield job
        elif status in ("failed", "timeout"):
            logger.warning(f"Job {job.job_id} {status}: {job.error_message}")
        else:
            jobs.append(job)
            logger.debug(f"Job {job.job_id} {status}, will check again later")
            # Sleep for a bit to avoid hammering the API
            time.sleep(10)


def create_annotations(
    job: ContextualizationJob,
    annotation_space: str,
    source: dm.DirectRelationReference,
    parameters: Parameters,
    logger: CogniteFunctionLogger,
) -> list[CogniteAnnotationApply]:
    annotation_list: list[CogniteAnnotationApply] = []
    for match_raw in job.result["items"]:
        match = MatchResult.load(match_raw)
        if (best_match := match.best_match) is None:
            logger.debug(f"No match found for {match.source.node_id!r}")
            continue
        source_id = match.source.node_id
        target_id = best_match.target.node_id
        now = datetime.now(timezone.utc).replace(microsecond=0)
        score = best_match.score
        status: Literal["Approved", "Suggested", "Rejected"] = "Suggested"
        if score >= parameters.auto_approval_threshold:
            status = "Approved"
        elif score <= parameters.auto_reject_threshold:
            status = "Rejected"
        external_id = create_annotation_id(source_id, target_id, EDGE_TYPE)
        annotation = CogniteAnnotationApply(
            space=annotation_space,
            external_id=external_id,
            start_node=(source_id.space, source_id.external_id),
            end_node=(target_id.space, target_id.external_id),
            type=(annotation_space, EDGE_TYPE),
            confidence=score,
            status=status,
            source=source,
            source_created_time=now,
            source_updated_time=now,
            source_created_user=FUNCTION_ID,
            source_updated_user=FUNCTION_ID,
            source_context=json.dumps({"end": best_match.target.view.dump(), "start": match.source.view.dump()}),
        )
        annotation_list.append(annotation)
    return annotation_list


def write_annotations(
    annotation_list: list[CogniteAnnotationApply],
    client: CogniteClient,
    job_id: int,
    logger: CogniteFunctionLogger,
) -> None:
    created = client.data_modeling.instances.apply(edges=annotation_list).edges

    create_count = sum(
        [1 for result in created if result.was_modified and result.created_time == result.last_updated_time]
    )
    update_count = sum(
        [1 for result in created if result.was_modified and result.created_time != result.last_updated_time]
    )
    unchanged_count = len(created) - create_count - update_count
    logger.info(
        f"Created {create_count} updated {update_count}, and {unchanged_count} unchanged annotations for {job_id}"
    )


def create_annotation_id(start: dm.NodeId, end: dm.NodeId, type: str) -> str:
    naive = f"{start.space}:{start.external_id}:{end.space}:{end.external_id}:{type}"
    if len(naive) < EXTERNAL_ID_LIMIT:
        return naive
    full_hash = sha256(naive.encode()).hexdigest()[:10]
    prefix = f"{start.external_id}:{end.external_id}:{type}"
    shorten = f"{prefix}:{full_hash}"
    if len(shorten) < EXTERNAL_ID_LIMIT:
        return shorten
    return prefix[: EXTERNAL_ID_LIMIT - 10] + full_hash


def load_config(client: CogniteClient, logger: CogniteFunctionLogger) -> Config:
    raw_config = client.extraction_pipelines.config.retrieve(EXTRACTION_PIPELINE_EXTERNAL_ID)
    if raw_config.config is None:
        raise ValueError("No config found for extraction pipeline")
    try:
        return Config.model_validate(yaml.safe_load(raw_config.config))
    except ValueError as e:
        logger.error(f"Invalid config: {e}")
        raise e



import contextlib
import io
from pathlib import Path
from unittest.mock import MagicMock

from cognite.client.data_classes.workflows import WorkflowScheduledTriggerRule, WorkflowTrigger

from cognite_toolkit._cdf_tk.loaders import ResourceWorker, WorkflowTriggerLoader
from tests.test_unit.approval_client import ApprovalToolkitClient


class TestResourceWorker:
    def test_mask_sensitive_data(self, toolkit_client_approval: ApprovalToolkitClient) -> None:
        toolkit_client_approval.append(
            WorkflowTrigger,
            WorkflowTrigger(
                "my_trigger",
                WorkflowScheduledTriggerRule(cron_expression="* * * * *"),
                "my_workflow",
                "v1",
                metadata={
                    WorkflowTriggerLoader._MetadataKey.secret_hash: "outdated-hash",
                },
            ),
        )
        loader = WorkflowTriggerLoader.create_loader(toolkit_client_approval.mock_client)

        worker = ResourceWorker(loader)
        local_file = MagicMock(spec=Path)
        local_file.read_text.return_value = """externalId: my_trigger
triggerRule:
  triggerType: schedule
  cronExpression: '* * * * *'
workflowExternalId: my_workflow
workflowVersion: v1
authentication:
  clientId: my_client_id
  clientSecret: my_super_secret_42
"""
        output_capture = io.StringIO()
        with contextlib.redirect_stdout(output_capture):
            _ = worker.prepare_resources([local_file], environment_variables={}, is_dry_run=False, verbose=True)

        terminal_output = output_capture.getvalue()
        assert "my_super_secret_42" not in terminal_output



from __future__ import annotations

from collections.abc import Iterator, Sequence
from typing import overload

from cognite.client._api_client import APIClient
from cognite.client.utils.useful_types import SequenceNotStr

from cognite_toolkit._cdf_tk.client.data_classes.agents import Agent, AgentList, AgentWrite


class AgentsAPI(APIClient):
    _RESOURCE_PATH = "/ai/agents"

    @overload
    def __call__(self) -> Iterator[Agent]: ...

    @overload
    def __call__(self, chunk_size: int) -> Iterator[AgentList]: ...

    def __call__(self, chunk_size: int | None = None) -> Iterator[Agent] | Iterator[AgentList]:
        """Iterate over agents.

        Args:
            chunk_size: The number of agents to return in each chunk. None will return all agents.

        Yields:
            Agent or AgentList

        """
        return self._list_generator(method="GET", resource_cls=Agent, list_cls=AgentList, chunk_size=chunk_size)

    def __iter__(self) -> Iterator[Agent]:
        return self.__call__()

    @overload
    def apply(self, agent: AgentWrite) -> Agent: ...

    @overload
    def apply(self, agent: Sequence[AgentWrite]) -> AgentList: ...

    def apply(self, agent: AgentWrite | Sequence[AgentWrite]) -> Agent | AgentList:
        """Create a new agent.

        Args:
            agent: AgentWrite or list of AgentWrite.

        Returns:
            Agent object.

        """
        headers = {"cdf-version": "alpha"}
        return self._create_multiple(
            list_cls=AgentList,
            resource_cls=Agent,
            items=agent,
            input_resource_cls=AgentWrite,
            headers=headers,
        )

    def retrieve(self, external_id: SequenceNotStr[str]) -> AgentList:
        """Retrieve an agent.

        Args:
            data_set_id: Data set id of the agent.

        Returns:
            Agent object.

        """
        body = self._create_body(external_id, True)

        headers = {"cdf-version": "alpha"}
        self._cognite_client.config.debug = True

        res = self._post(url_path=self._RESOURCE_PATH + "/byids", json=body, headers=headers)
        return AgentList._load(res.json()["items"], cognite_client=self._cognite_client)

    @staticmethod
    def _create_body(external_id: str | SequenceNotStr[str], ignore_unknown_ids: bool = False) -> dict:
        ids = [external_id] if isinstance(external_id, str) else external_id
        body = {"items": [{"externalId": external_id} for external_id in ids], "ignoreUnknownIds": ignore_unknown_ids}
        return body

    def delete(self, external_id: str | SequenceNotStr[str]) -> None:
        """Delete an agent.

        Args:
            external_id: External ID of the agent.

        Returns:
            None

        """
        body = self._create_body(external_id)
        headers = {"cdf-version": "alpha"}

        self._post(url_path=self._RESOURCE_PATH + "/delete", json=body, headers=headers)

    def list(self) -> AgentList:
        """List agents.

        Returns:
            AgentList

        """
        return self._list(method="GET", resource_cls=Agent, list_cls=AgentList)



from __future__ import annotations

from typing import Any

import pytest

from cognite_toolkit._cdf_tk._parameters import ANY_INT, ANY_STR


class TestAnyStrInt:
    @pytest.mark.parametrize("other", [1, "a", None])
    def test_any_str_equals(self, other: Any) -> None:
        is_string = isinstance(other, str)
        is_equal = ANY_STR == other

        assert is_string == is_equal

    @pytest.mark.parametrize("other", [1, "a", None])
    def test_any_int_equals(self, other: Any) -> None:
        is_int = isinstance(other, int)
        is_equal = ANY_INT == other

        assert is_int == is_equal

    @pytest.mark.parametrize(
        "dump, spec, expected",
        [
            (("metadata",), ("metadata", ANY_STR), False),
            (("metadata",), ("metadata", ANY_INT), False),
            (("metadata", 0), ("metadata", ANY_INT), True),
            (("metadata",), ("metadata",), True),
            (("metadata",), ("metadata", "a"), False),
            (("metadata", "a"), ("metadata", ANY_STR), True),
        ],
    )
    def test_any_str_tuple_equals(self, dump: tuple, spec: tuple, expected: bool) -> None:
        actual = dump == spec

        assert actual == expected

    def test_less_than(self) -> None:
        is_less_than = ANY_INT < "str"
        assert is_less_than is True

    def test_less_than_reverse(self) -> None:
        is_less_than = "str" < ANY_INT
        assert is_less_than is False



from collections.abc import Iterable
from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.resource_classes import FunctionsYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.test_unit.utils import find_resources


def invalid_function_test_cases() -> Iterable:
    yield pytest.param(
        {"name": "my_function"},
        {"Missing required field: 'externalId'"},
        id="Missing externalId",
    )
    yield pytest.param(
        {
            "externalId": "my_function",
            "name": "my_function",
            "runtime": "py36",
            "secrets": {f"secret_name{i}": f"super_secret{i}" for i in range(31)},
        },
        {
            "In field runtime input should be 'py39', 'py310' or 'py311'. Got 'py36'.",
            "In field secrets dictionary should have at most 30 items after validation, not 31",
        },
        id="Invalid runtime",
    )


class TestFunctionsYAML:
    @pytest.mark.parametrize("data", list(find_resources("function")))
    def test_load_valid_function(self, data: dict[str, object]) -> None:
        loaded = FunctionsYAML.model_validate(data)

        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data, expected_errors", list(invalid_function_test_cases()))
    def test_invalid_function_error_messages(self, data: dict | list, expected_errors: set[str]) -> None:
        """Test the validate_resource_yaml function for GroupYAML."""
        warning_list = validate_resource_yaml_pydantic(data, FunctionsYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)

        assert set(format_warning.errors) == expected_errors



from pathlib import Path
from typing import Annotated, Any, Optional

import typer
from rich import print

from cognite_toolkit._cdf_tk.cdf_toml import CDFToml
from cognite_toolkit._cdf_tk.commands import (
    RunFunctionCommand,
    RunTransformationCommand,
    RunWorkflowCommand,
)
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables

CDF_TOML = CDFToml.load(Path.cwd())


class RunApp(typer.Typer):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.callback(invoke_without_command=True)(self.main)
        self.command("transformation")(self.run_transformation)
        self.command("workflow")(self.run_workflow)
        self.add_typer(RunFunctionApp(*args, **kwargs), name="function")

    @staticmethod
    def main(ctx: typer.Context) -> None:
        """Commands to execute processes in CDF."""
        if ctx.invoked_subcommand is None:
            print("Use [bold yellow]cdf run --help[/] for more information.")

    @staticmethod
    def run_transformation(
        ctx: typer.Context,
        external_id: Annotated[
            str,
            typer.Option(
                "--external-id",
                "-e",
                prompt=True,
                help="External id of the transformation to run.",
            ),
        ],
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will run the specified transformation using a one-time session."""
        cmd = RunTransformationCommand()
        client = EnvironmentVariables.create_from_environment().get_client()
        cmd.run(lambda: cmd.run_transformation(client, external_id))

    @staticmethod
    def run_workflow(
        ctx: typer.Context,
        external_id: Annotated[
            Optional[str],
            typer.Option(
                "--external-id",
                "-e",
                help="External id of the workflow to run. If not provided, you will be prompted to select one.",
            ),
        ] = None,
        version: Annotated[
            Optional[str],
            typer.Option(
                "--version",
                "-v",
                help="Version of the workflow to run. If not provided, the first found version will be used.",
            ),
        ] = None,
        organization_dir: Annotated[
            Path,
            typer.Option(
                "--organization-dir",
                "-o",
                help="Path to project directory with the modules. This is used to search for available functions.",
            ),
        ] = CDF_TOML.cdf.default_organization_dir,
        env_name: Annotated[
            Optional[str],
            typer.Option(
                "--env",
                "-e",
                help="Name of the build environment to use. If not provided, the default environment will be used.",
            ),
        ] = CDF_TOML.cdf.default_env,
        wait: Annotated[
            bool,
            typer.Option(
                "--wait",
                "-w",
                help="Whether to wait for the workflow to complete.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will run the specified workflow."""
        cmd = RunWorkflowCommand()
        env_vars = EnvironmentVariables.create_from_environment()
        cmd.run(lambda: cmd.run_workflow(env_vars, organization_dir, env_name, external_id, version, wait))


class RunFunctionApp(typer.Typer):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.callback(invoke_without_command=True)(self.main)
        self.command("local")(self.run_local)
        self.command("live")(self.run_cdf)

    @staticmethod
    def main(ctx: typer.Context) -> None:
        """Commands to execute function."""
        if ctx.invoked_subcommand is None:
            print("Use [bold yellow]cdf run function --help[/] for more information.")

    @staticmethod
    def run_local(
        ctx: typer.Context,
        external_id: Annotated[
            Optional[str],
            typer.Argument(
                help="External id of the function to run.",
            ),
        ] = None,
        organization_dir: Annotated[
            Path,
            typer.Option(
                "--organization-dir",
                "-o",
                help="Path to project directory with the modules. This is used to search for available functions.",
            ),
        ] = CDF_TOML.cdf.default_organization_dir,
        env_name: Annotated[
            Optional[str],
            typer.Option(
                "--env",
                "-e",
                help="Name of the build environment to use. If not provided, the default environment will be used.",
            ),
        ] = CDF_TOML.cdf.default_env,
        schedule: Annotated[
            Optional[str],
            typer.Option(
                "--schedule",
                "-s",
                help="Schedule to run the function with (if any). The data and credentials "
                "will be taken from the schedule. If the schedule has no credentials, the "
                "default from the environment will be used.",
            ),
        ] = None,
        rebuild_env: Annotated[
            bool,
            typer.Option(
                "--rebuild-env",
                help="Whether to rebuild the environment before running the function.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will run the specified function locally."""
        cmd = RunFunctionCommand()
        env_vars = EnvironmentVariables.create_from_environment()
        cmd.run(
            lambda: cmd.run_local(
                env_vars,
                organization_dir,
                env_name,
                external_id,
                schedule,
                rebuild_env,
            )
        )

    @staticmethod
    def run_cdf(
        ctx: typer.Context,
        external_id: Annotated[
            Optional[str],
            typer.Argument(
                help="External id of the function to run. If not provided, the function "
                "will be selected interactively.",
            ),
        ] = None,
        organization_dir: Annotated[
            Path,
            typer.Option(
                "--organization-dir",
                "-o",
                help="Path to organization directory with the modules. This is used to search for available functions.",
            ),
        ] = CDF_TOML.cdf.default_organization_dir,
        env_name: Annotated[
            Optional[str],
            typer.Option(
                "--env",
                "-e",
                help="Name of the build environment to use. If not provided, the default environment will be used.",
            ),
        ] = CDF_TOML.cdf.default_env,
        schedule: Annotated[
            Optional[str],
            typer.Option(
                "--schedule",
                "-s",
                help="The name of the schedule to pick the data from",
            ),
        ] = None,
        wait: Annotated[
            bool,
            typer.Option(
                "--wait",
                "-w",
                help="Whether to wait for the function to complete.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will run the specified function (assuming it is deployed) in CDF."""
        cmd = RunFunctionCommand()
        env_vars = EnvironmentVariables.create_from_environment()
        cmd.run(lambda: cmd.run_cdf(env_vars, organization_dir, env_name, external_id, schedule, wait))



from __future__ import annotations

import sys
from collections.abc import Mapping
from graphlib import CycleError
from pathlib import Path

from yaml import YAMLError

if sys.version_info >= (3, 11):
    from tomllib import TOMLDecodeError
else:
    from tomli import TOMLDecodeError


class ToolkitError(Exception):
    def __repr__(self) -> str:
        # Repr is what is called by rich when the exception is printed.
        return str(self)


class ToolkitNotSupported(ToolkitError): ...


class ToolkitInvalidSettingsError(ToolkitError):
    pass


class ToolkitMigrationError(ToolkitError):
    pass


class ToolkitVersionError(ToolkitError):
    pass


class ToolkitModuleVersionError(ToolkitError):
    pass


class ToolkitEnvError(ToolkitError):
    pass


class ToolkitMissingResourceError(ToolkitError):
    pass


class ToolkitInvalidFunctionError(ToolkitError): ...


class ToolkitCleanResourceError(ToolkitError):
    pass


class ToolkitDeployResourceError(ToolkitError):
    pass


class ToolkitMissingModuleError(ToolkitError):
    pass


class ToolkitMissingModulesError(ToolkitError):
    pass


class ToolkitDuplicatedResourceError(ToolkitError):
    pass


class ToolkitDuplicatedModuleError(ToolkitError):
    def __init__(self, message: str, duplicated: dict[str, list[Path]]) -> None:
        super().__init__(message)
        self.duplicated = duplicated

    def __str__(self) -> str:
        lines = [super().__str__()]
        for module_name, paths in self.duplicated.items():
            locations = "\n        ".join(sorted(path.as_posix() for path in paths))
            lines.append(f"    {module_name} exists in:\n        {locations}")
        lines.append(
            "    You can use the path syntax to disambiguate between modules with the same name. For example "
            "'cognite_modules/core/cdf_apm_base' instead of 'cdf_apm_base'."
        )
        return "\n".join(lines)

    def __repr__(self) -> str:
        # Repr is what is called by rich when the exception is printed.
        return str(self)


class ToolkitNotADirectoryError(NotADirectoryError, ToolkitError):
    pass


class ToolkitIsADirectoryError(IsADirectoryError, ToolkitError):
    pass


class ToolkitFileNotFoundError(FileNotFoundError, ToolkitError):
    pass


class ToolkitFileExistsError(FileExistsError, ToolkitError):
    pass


class ToolkitValidationError(ToolkitError):
    pass


class ToolkitCycleError(CycleError, ToolkitValidationError):
    def __str__(self) -> str:
        return self.args[0]


class ToolkitYAMLFormatError(YAMLError, ToolkitValidationError):
    pass


class ToolkitTOMLFormatError(TOMLDecodeError, ToolkitValidationError): ...


class ToolkitInvalidParameterNameError(ToolkitValidationError):
    def __init__(self, message: str, identifier: str, correct_by_wrong_parameter: Mapping[str, str | None]) -> None:
        super().__init__(message)
        self.identifier = identifier
        self.parameter = correct_by_wrong_parameter

    def __str__(self) -> str:
        parameters = []
        for wrong, correct in self.parameter.items():
            if correct is not None:
                parameters.append(f"{wrong!r} should be {correct!r}")
            else:
                parameters.append(f"{wrong!r} is invalid")
        parameter_str = "    \n".join(parameters)
        message = super().__str__()
        return f"{message}\nIn {self.identifier!r} the following parameters are invalid: {parameter_str}"

    def __repr__(self) -> str:
        # Repr is what is called by rich when the exception is printed.
        return str(self)


class ToolkitValueError(ValueError, ToolkitError):
    pass


class ToolkitMissingValueError(ToolkitValueError):
    pass


class ToolkitKeyError(KeyError, ToolkitError): ...


class ToolkitTypeError(TypeError, ToolkitError): ...


class ToolkitRequiredValueError(ToolkitError, ValueError):
    pass


class ToolkitResourceMissingError(ToolkitError):
    def __init__(self, message: str, resource: str) -> None:
        super().__init__(message)
        self.resource = resource

    def __str__(self) -> str:
        return f"{super().__str__()}\nResource {self.resource!r} is missing"


class ToolkitWrongResourceError(RuntimeError):
    """Special exception used by a loader to signal that the resource is not the expected type."""

    ...


class UploadFileError(ToolkitError):
    pass


class ResourceRetrievalError(ToolkitError): ...


class ResourceCreationError(ToolkitError):
    pass


class ResourceDeleteError(ToolkitError): ...


class ResourceUpdateError(ToolkitError):
    pass


class AmbiguousResourceFileError(ToolkitError):
    pass


class AuthenticationError(ToolkitError):
    pass


class AuthorizationError(ToolkitError):
    pass


class GraphQLParseError(ToolkitError): ...


class CDFAPIError(ToolkitError, RuntimeError):
    """Error raised when the CDF API returns an error."""

    ...


class ToolkitMissingDependencyError(ToolkitError):
    """Error raised when a required dependency is missing."""

    def __init__(self, message: str) -> None:
        super().__init__(message)



from __future__ import annotations

import dataclasses
import typing
from collections.abc import Hashable, Iterable, MutableSet
from dataclasses import dataclass
from functools import total_ordering

# Ruff wants to get AbstractSet from collections.abc, but mypy does not allow it.
from typing import AbstractSet, Generic, TypeVar, final  # noqa: UP035

from cognite.client.utils._text import to_camel_case

from .constants import ANYTHING, SINGLETONS


@total_ordering
@dataclass(frozen=True)
class Parameter:
    path: tuple[str | int, ...]

    @property
    def key(self) -> str:
        if isinstance(self.path[-1], str):
            return self.path[-1]
        else:
            return str(self.path[-1])

    def __lt__(self, other: Parameter) -> bool:
        if not isinstance(other, Parameter):
            return NotImplemented
        return self.path < other.path

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Parameter):
            return NotImplemented
        return self.path == other.path

    @property
    def has_any_type(self) -> bool:
        return any(name in SINGLETONS for name in self.path)


@final
@dataclass(frozen=True)
class ParameterSpec(Parameter):
    types: frozenset[str]
    is_required: bool
    _is_nullable: bool | None = None

    @property
    def is_nullable(self) -> bool:
        return self._is_nullable or not self.is_required

    # The eq, ne, and hash methods must be implemented for each subclass to get the set operations to work correctly.
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Parameter):
            return NotImplemented
        return self.path == other.path

    def __ne__(self, other: object) -> bool:
        if not isinstance(other, Parameter):
            return NotImplemented
        return self.path != other.path

    def __hash__(self) -> int:
        return hash(self.path)


@final
@dataclass(frozen=True)
class ParameterValue(Parameter):
    type: str
    value: str | int | float | bool | None

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Parameter):
            return NotImplemented
        return self.path == other.path

    def __ne__(self, other: object) -> bool:
        if not isinstance(other, Parameter):
            return NotImplemented
        return self.path != other.path

    def __hash__(self) -> int:
        return hash(self.path)


T_Parameter = TypeVar("T_Parameter", bound=Parameter)


class ParameterSet(Hashable, MutableSet, Generic[T_Parameter]):
    def __init__(self, iterable: Iterable[T_Parameter] = ()) -> None:
        self.data: set[T_Parameter] = set(iterable)

    def as_camel_case(self) -> ParameterSet[T_Parameter]:
        output = type(self)()
        for parameter in self:
            new_path = tuple(
                to_camel_case(name) if name not in SINGLETONS and isinstance(name, str) else name
                for name in parameter.path
            )
            # This is because parameters are immutable
            new_param_copy = dataclasses.replace(parameter, path=new_path)
            output.add(new_param_copy)
        return output

    @property
    def has_any_type(self) -> bool:
        return any(parameter.has_any_type for parameter in self)

    def subset_any_type_paths(self) -> ParameterSet[T_Parameter]:
        return type(self)(parameter for parameter in self if parameter.has_any_type)

    def subset_anything_paths(self) -> ParameterSet[T_Parameter]:
        return type(self)(parameter for parameter in self if parameter.path[-1] is ANYTHING)

    def subset(self, path: tuple[str | int, ...] | int) -> ParameterSet[T_Parameter]:
        if isinstance(path, tuple):
            return type(self)(parameter for parameter in self if parameter.path[: len(path)] == path)
        elif isinstance(path, int):
            return type(self)(parameter for parameter in self if len(parameter.path) <= path)
        raise TypeError(f"Expected tuple or int, got {type(path)}")

    def __hash__(self) -> int:
        return hash(self.data)

    def __contains__(self, value: object) -> bool:
        return value in self.data

    def __iter__(self) -> typing.Iterator[T_Parameter]:
        return iter(self.data)

    def __len__(self) -> int:
        return len(self.data)

    def __repr__(self) -> str:
        return repr(self.data)

    def add(self, item: T_Parameter) -> None:
        self.data.add(item)

    def discard(self, item: T_Parameter) -> None:
        self.data.discard(item)

    def update(self, other: ParameterSet[T_Parameter]) -> None:
        self.data.update(other.data)

    def difference(self, other: ParameterSet[T_Parameter]) -> ParameterSet[T_Parameter]:
        output = type(self)(self.data.difference(other.data))
        other_any = other.subset_any_type_paths()
        other_anything = other.subset_anything_paths()
        if not other_any:
            # Operation done using hashes
            return output
        # Operation done using equality as ANY_STR and ANY_INT does not match on hash
        # Note that this a Parameter with ANY_STR in the path can remove multiple items.
        discard = set()
        for item in output:
            for any_ in other_any:
                if item == any_:
                    discard.add(item)
            for anything in other_anything:
                if item.path[: len(anything.path)] == anything.path:
                    discard.add(item)
        for d in discard:
            output.discard(d)
        return output

    def __sub__(self, other: AbstractSet) -> ParameterSet[T_Parameter]:
        return self.difference(type(self)(other))


class ParameterSpecSet(ParameterSet[ParameterSpec]):
    def __init__(self, iterable: Iterable[ParameterSpec] = (), spec_name: str | None = None) -> None:
        super().__init__(iterable)
        self.is_complete = True
        self.spec_name = spec_name

    def required(self, level: int | None = None) -> ParameterSet[ParameterSpec]:
        if level is None:
            return ParameterSet[ParameterSpec](parameter for parameter in self if parameter.is_required)
        else:
            return ParameterSet[ParameterSpec](
                parameter for parameter in self if len(parameter.path) <= level and parameter.is_required
            )

    def as_camel_case(self) -> ParameterSpecSet:
        output = typing.cast(ParameterSpecSet, super().as_camel_case())
        output.is_complete = self.is_complete
        return output

    def update(self, other: ParameterSet[ParameterSpec]) -> None:
        if isinstance(other, ParameterSpecSet):
            self.is_complete &= other.is_complete
        super().update(other)



import os
from collections.abc import Hashable
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest
from _pytest.monkeypatch import MonkeyPatch
from cognite.client.data_classes import ExtractionPipelineConfig

from cognite_toolkit._cdf_tk.client.data_classes.raw import RawDatabase, RawTable
from cognite_toolkit._cdf_tk.commands import CleanCommand
from cognite_toolkit._cdf_tk.loaders import (
    DataSetsLoader,
    ExtractionPipelineConfigLoader,
    ExtractionPipelineLoader,
    RawDatabaseLoader,
    RawTableLoader,
    ResourceLoader,
    ResourceWorker,
)
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables
from tests.test_unit.approval_client import ApprovalToolkitClient


class TestExtractionPipelineDependencies:
    _yaml = """
        externalId: 'ep_src_asset_hamburg_sap'
        name: 'Hamburg SAP'
        dataSetId: 12345
    """

    config_yaml = """
        externalId: 'ep_src_asset'
        description: 'DB extractor config reading data from Springfield SAP'
    """

    def test_load_extraction_pipeline_upsert_update_one(
        self, toolkit_client_approval: ApprovalToolkitClient, monkeypatch: MonkeyPatch
    ) -> None:
        toolkit_client_approval.append(
            ExtractionPipelineConfig,
            ExtractionPipelineConfig(
                external_id="ep_src_asset",
                description="DB extractor config reading data from Springfield SAP",
                config="\n    logger: \n        {level: WARN}",
            ),
        )

        local_file = MagicMock(spec=Path)
        local_file.read_text.return_value = self.config_yaml

        loader = ExtractionPipelineConfigLoader.create_loader(toolkit_client_approval.mock_client)
        worker = ResourceWorker(loader)
        to_create, changed, to_delete, unchanged = worker.prepare_resources([local_file])
        assert {
            "create": len(to_create),
            "changed": len(changed),
            "delete": len(to_delete),
            "unchanged": len(unchanged),
        } == {"create": 0, "changed": 1, "delete": 0, "unchanged": 0}

    def test_load_extraction_pipeline_delete_one(
        self,
        toolkit_client_approval: ApprovalToolkitClient,
        env_vars_with_client: EnvironmentVariables,
        monkeypatch: MonkeyPatch,
    ) -> None:
        toolkit_client_approval.append(
            ExtractionPipelineConfig,
            ExtractionPipelineConfig(
                external_id="ep_src_asset",
                description="DB extractor config reading data from Springfield SAP",
                config="\n    logger: \n        {level: WARN}",
            ),
        )

        local_file = MagicMock(spec=Path)
        local_file.read_text.return_value = self.config_yaml
        local_file.stem = "ep_src_asset"

        cmd = CleanCommand(print_warning=False)
        loader = ExtractionPipelineConfigLoader.create_loader(env_vars_with_client.get_client())
        with patch.object(ExtractionPipelineConfigLoader, "find_files", return_value=[local_file]):
            res = cmd.clean_resources(loader, env_vars_with_client, [], dry_run=True, drop=True)
            assert res is not None
            assert res.deleted == 1


class TestExtractionPipelineLoader:
    @pytest.mark.parametrize(
        "item, expected",
        [
            pytest.param(
                {
                    "dataSetExternalId": "ds_my_dataset",
                    "rawTables": [
                        {"dbName": "my_db", "tableName": "my_table"},
                        {"dbName": "my_db", "tableName": "my_table2"},
                    ],
                },
                [
                    (DataSetsLoader, "ds_my_dataset"),
                    (RawDatabaseLoader, RawDatabase("my_db")),
                    (RawTableLoader, RawTable("my_db", "my_table")),
                    (RawTableLoader, RawTable("my_db", "my_table2")),
                ],
                id="Extraction pipeline to Table",
            ),
        ],
    )
    def test_get_dependent_items(self, item: dict, expected: list[tuple[type[ResourceLoader], Hashable]]) -> None:
        actual = ExtractionPipelineLoader.get_dependent_items(item)

        assert list(actual) == expected

    @patch.dict(
        os.environ,
        {
            "INGESTION_CLIENT_ID": "this-is-the-ingestion-client-id",
            "INGESTION_CLIENT_SECRET": "this-is-the-ingestion-client-secret",
            "NON-SECRET": "this-is-not-a-secret",
        },
    )
    def test_omit_environment_variables(
        self, env_vars_with_client: EnvironmentVariables, monkeypatch: MonkeyPatch
    ) -> None:
        local_file = MagicMock(spec=Path)
        local_file.read_text.return_value = """
            - externalId: 'ep_src_asset'
              name: 'Hamburg SAP'
              config: 'secret: ${INGESTION_CLIENT_SECRET}'
            - externalId: 'ep_src_asset_2'
              name: '${NON-SECRET}'
              config: 'secret: ${INGESTION_CLIENT_SECRET}'
        """
        local_file.stem = "ep_src_asset"

        loader = ExtractionPipelineConfigLoader.create_loader(env_vars_with_client.get_client())
        res = loader.load_resource_file(filepath=local_file, environment_variables=env_vars_with_client.dump())
        # Assert that env vars are skipped for this loader
        assert res[0]["config"] == "secret: ${INGESTION_CLIENT_SECRET}"
        assert res[1]["name"] == "this-is-not-a-secret"



from __future__ import annotations

import json
from collections.abc import Callable, Hashable, Iterable, Sequence
from contextlib import suppress
from typing import Any

from cognite.client.data_classes import capabilities
from cognite.client.data_classes._base import T_CogniteResourceList
from cognite.client.data_classes.capabilities import Capability
from cognite.client.exceptions import CogniteAPIError
from cognite.client.utils.useful_types import SequenceNotStr

from cognite_toolkit._cdf_tk.client.data_classes.robotics import (
    DataPostProcessing,
    DataPostProcessingList,
    DataPostProcessingWrite,
    DataPostProcessingWriteList,
    Frame,
    FrameList,
    FrameWrite,
    FrameWriteList,
    Location,
    LocationList,
    LocationWrite,
    LocationWriteList,
    Map,
    MapList,
    MapWrite,
    MapWriteList,
    RobotCapability,
    RobotCapabilityList,
    RobotCapabilityWrite,
    RobotCapabilityWriteList,
)
from cognite_toolkit._cdf_tk.loaders._base_loaders import ResourceLoader
from cognite_toolkit._cdf_tk.utils.diff_list import diff_list_hashable


class RoboticFrameLoader(ResourceLoader[str, FrameWrite, Frame, FrameWriteList, FrameList]):
    folder_name = "robotics"
    filename_pattern = r"^.*\.Frame$"  # Matches all yaml files whose stem ends with '.Frame'.
    resource_cls = Frame
    resource_write_cls = FrameWrite
    list_cls = FrameList
    list_write_cls = FrameWriteList
    kind = "Frame"
    _doc_url = "Frames/operation/createFrames"

    @property
    def display_name(self) -> str:
        return "robotics frames"

    @classmethod
    def get_id(cls, item: Frame | FrameWrite | dict) -> str:
        if isinstance(item, dict):
            return item["externalId"]
        if not item.external_id:
            raise KeyError("Frame must have external_id")
        return item.external_id

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"externalId": id}

    @classmethod
    def get_required_capability(
        cls, items: Sequence[FrameWrite] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []
        return capabilities.RoboticsAcl(
            [
                capabilities.RoboticsAcl.Action.Read,
                capabilities.RoboticsAcl.Action.Create,
                capabilities.RoboticsAcl.Action.Delete,
                capabilities.RoboticsAcl.Action.Update,
                capabilities.RoboticsAcl.Action.Delete,
            ],
            capabilities.RoboticsAcl.Scope.All(),
        )

    def create(self, items: FrameWriteList) -> FrameList:
        return self.client.robotics.frames.create(items)

    def retrieve(self, ids: SequenceNotStr[str]) -> FrameList:
        return _fallback_to_one_by_one(self.client.robotics.frames.retrieve, ids, FrameList)

    def update(self, items: FrameWriteList) -> FrameList:
        return self.client.robotics.frames.update(items)

    def delete(self, ids: SequenceNotStr[str]) -> int:
        try:
            self.client.robotics.frames.delete(ids)
        except CogniteAPIError as e:
            return len(e.successful)
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[Frame]:
        return iter(self.client.robotics.frames)


class RoboticLocationLoader(ResourceLoader[str, LocationWrite, Location, LocationWriteList, LocationList]):
    folder_name = "robotics"
    filename_pattern = r"^.*\.Location$"  # Matches all yaml files whose stem ends with '.Location'.
    resource_cls = Location
    resource_write_cls = LocationWrite
    list_cls = LocationList
    list_write_cls = LocationWriteList
    kind = "Location"
    _doc_url = "Locations/operation/createLocations"

    @property
    def display_name(self) -> str:
        return "robotics locations"

    @classmethod
    def get_id(cls, item: Location | LocationWrite | dict) -> str:
        if isinstance(item, dict):
            return item["externalId"]
        if not item.external_id:
            raise KeyError("Location must have external_id")
        return item.external_id

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"externalId": id}

    @classmethod
    def get_required_capability(
        cls, items: Sequence[LocationWrite] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []
        actions = (
            [
                capabilities.RoboticsAcl.Action.Read,
            ]
            if read_only
            else [
                capabilities.RoboticsAcl.Action.Read,
                capabilities.RoboticsAcl.Action.Create,
                capabilities.RoboticsAcl.Action.Delete,
                capabilities.RoboticsAcl.Action.Update,
                capabilities.RoboticsAcl.Action.Delete,
            ]
        )

        return capabilities.RoboticsAcl(actions, capabilities.RoboticsAcl.Scope.All())

    def create(self, items: LocationWriteList) -> LocationList:
        return self.client.robotics.locations.create(items)

    def retrieve(self, ids: SequenceNotStr[str]) -> LocationList:
        return _fallback_to_one_by_one(self.client.robotics.locations.retrieve, ids, LocationList)

    def update(self, items: LocationWriteList) -> LocationList:
        return self.client.robotics.locations.update(items)

    def delete(self, ids: SequenceNotStr[str]) -> int:
        try:
            self.client.robotics.locations.delete(ids)
        except CogniteAPIError as e:
            return len(e.successful)
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[Location]:
        return iter(self.client.robotics.locations)


class RoboticsDataPostProcessingLoader(
    ResourceLoader[
        str, DataPostProcessingWrite, DataPostProcessing, DataPostProcessingWriteList, DataPostProcessingList
    ]
):
    folder_name = "robotics"
    filename_pattern = r"^.*\.DataPostProcessing$"  # Matches all yaml files whose stem ends with '.DataPostProcessing'.
    resource_cls = DataPostProcessing
    resource_write_cls = DataPostProcessingWrite
    list_cls = DataPostProcessingList
    list_write_cls = DataPostProcessingWriteList
    kind = "DataPostProcessing"
    _doc_url = "DataPostProcessing/operation/createDataPostProcessing"

    @property
    def display_name(self) -> str:
        return "robotics data postprocessing"

    @classmethod
    def get_id(cls, item: DataPostProcessing | DataPostProcessingWrite | dict) -> str:
        if isinstance(item, dict):
            return item["externalId"]
        if not item.external_id:
            raise KeyError("DataPostProcessing must have external_id")
        return item.external_id

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"externalId": id}

    @classmethod
    def get_required_capability(
        cls, items: Sequence[DataPostProcessingWrite] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []
        actions = (
            [
                capabilities.RoboticsAcl.Action.Read,
            ]
            if read_only
            else [
                capabilities.RoboticsAcl.Action.Read,
                capabilities.RoboticsAcl.Action.Create,
                capabilities.RoboticsAcl.Action.Delete,
                capabilities.RoboticsAcl.Action.Update,
                capabilities.RoboticsAcl.Action.Delete,
            ]
        )

        return capabilities.RoboticsAcl(actions, capabilities.RoboticsAcl.Scope.All())

    def create(self, items: DataPostProcessingWriteList) -> DataPostProcessingList:
        return self.client.robotics.data_postprocessing.create(items)

    def retrieve(self, ids: SequenceNotStr[str]) -> DataPostProcessingList:
        return _fallback_to_one_by_one(self.client.robotics.data_postprocessing.retrieve, ids, DataPostProcessingList)

    def update(self, items: DataPostProcessingWriteList) -> DataPostProcessingList:
        # There is a bug in the /update endpoint that requires the input_schema to be a string
        # and not an object. This is a workaround until the bug is fixed.
        # We do the serialization to avoid modifying the original object.
        to_update = []
        for item in items:
            if isinstance(item.input_schema, dict):
                update = DataPostProcessingWrite.load(item.dump())
                update.input_schema = json.dumps(item.input_schema)  # type: ignore[assignment]
                to_update.append(update)
            else:
                to_update.append(item)

        return self.client.robotics.data_postprocessing.update(to_update)

    def delete(self, ids: SequenceNotStr[str]) -> int:
        try:
            self.client.robotics.data_postprocessing.delete(ids)
        except CogniteAPIError as e:
            return len(e.successful)
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[DataPostProcessing]:
        return iter(self.client.robotics.data_postprocessing)

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path[0] == "inputSchema":
            return diff_list_hashable(local, cdf)
        return super().diff_list(local, cdf, json_path)


class RobotCapabilityLoader(
    ResourceLoader[str, RobotCapabilityWrite, RobotCapability, RobotCapabilityWriteList, RobotCapabilityList]
):
    folder_name = "robotics"
    filename_pattern = r"^.*\.RobotCapability$"  # Matches all yaml files whose stem ends with '.RobotCapability'.
    resource_cls = RobotCapability
    resource_write_cls = RobotCapabilityWrite
    list_cls = RobotCapabilityList
    list_write_cls = RobotCapabilityWriteList
    kind = "RobotCapability"
    _doc_url = "RobotCapabilities/operation/createRobotCapabilities"

    @property
    def display_name(self) -> str:
        return "robotics robot capabilities"

    @classmethod
    def get_id(cls, item: RobotCapability | RobotCapabilityWrite | dict) -> str:
        if isinstance(item, dict):
            return item["externalId"]
        if not item.external_id:
            raise KeyError("RobotCapability must have external_id")
        return item.external_id

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"externalId": id}

    @classmethod
    def get_required_capability(
        cls, items: Sequence[RobotCapabilityWrite] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []
        actions = (
            [
                capabilities.RoboticsAcl.Action.Read,
            ]
            if read_only
            else [
                capabilities.RoboticsAcl.Action.Read,
                capabilities.RoboticsAcl.Action.Create,
                capabilities.RoboticsAcl.Action.Delete,
                capabilities.RoboticsAcl.Action.Update,
                capabilities.RoboticsAcl.Action.Delete,
            ]
        )

        return capabilities.RoboticsAcl(actions, capabilities.RoboticsAcl.Scope.All())

    def create(self, items: RobotCapabilityWriteList) -> RobotCapabilityList:
        return self.client.robotics.capabilities.create(items)

    def retrieve(self, ids: SequenceNotStr[str]) -> RobotCapabilityList:
        return _fallback_to_one_by_one(self.client.robotics.capabilities.retrieve, ids, RobotCapabilityList)

    def update(self, items: RobotCapabilityWriteList) -> RobotCapabilityList:
        # There is a bug in the /update endpoint that requires the input_schema to be a string
        # and not an object. This is a workaround until the bug is fixed.
        # We do the serialization to avoid modifying the original object.
        to_update = []
        for item in items:
            if isinstance(item.input_schema, dict) or isinstance(item.data_handling_schema, dict):
                update = RobotCapabilityWrite.load(item.dump())
                if isinstance(item.data_handling_schema, dict):
                    update.data_handling_schema = json.dumps(item.data_handling_schema)  # type: ignore[assignment]
                if isinstance(item.input_schema, dict):
                    update.input_schema = json.dumps(item.input_schema)  # type: ignore[assignment]
                to_update.append(update)
            else:
                to_update.append(item)

        return self.client.robotics.capabilities.update(to_update)

    def delete(self, ids: SequenceNotStr[str]) -> int:
        try:
            self.client.robotics.capabilities.delete(ids)
        except CogniteAPIError as e:
            return len(e.successful)
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[RobotCapability]:
        return iter(self.client.robotics.capabilities)

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path[0] in {"inputSchema", "dataHandlingSchema"}:
            return diff_list_hashable(local, cdf)
        return super().diff_list(local, cdf, json_path)


class RoboticMapLoader(ResourceLoader[str, MapWrite, Map, MapWriteList, MapList]):
    folder_name = "robotics"
    filename_pattern = r"^.*\.Map$"  # Matches all yaml files whose stem ends with '.Map'.
    resource_cls = Map
    resource_write_cls = MapWrite
    list_cls = MapList
    list_write_cls = MapWriteList
    kind = "Map"
    dependencies = frozenset({RoboticFrameLoader, RoboticLocationLoader})
    _doc_url = "Maps/operation/createMaps"

    @property
    def display_name(self) -> str:
        return "robotics maps"

    @classmethod
    def get_id(cls, item: Map | MapWrite | dict) -> str:
        if isinstance(item, dict):
            return item["externalId"]
        if not item.external_id:
            raise KeyError("Map must have external_id")
        return item.external_id

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"externalId": id}

    @classmethod
    def get_required_capability(
        cls, items: Sequence[MapWrite] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [
                capabilities.RoboticsAcl.Action.Read,
            ]
            if read_only
            else [
                capabilities.RoboticsAcl.Action.Read,
                capabilities.RoboticsAcl.Action.Create,
                capabilities.RoboticsAcl.Action.Delete,
                capabilities.RoboticsAcl.Action.Update,
                capabilities.RoboticsAcl.Action.Delete,
            ]
        )

        return capabilities.RoboticsAcl(actions, capabilities.RoboticsAcl.Scope.All())

    def dump_resource(self, resource: Map, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dump = resource.as_write().dump()
        local = local or {}
        if dump.get("scale") == 1.0 and "scale" not in local:
            # Default value set on the server side.
            del dump["scale"]
        return dump

    def create(self, items: MapWriteList) -> MapList:
        return self.client.robotics.maps.create(items)

    def retrieve(self, ids: SequenceNotStr[str]) -> MapList:
        return _fallback_to_one_by_one(self.client.robotics.maps.retrieve, ids, MapList)

    def update(self, items: MapWriteList) -> MapList:
        return self.client.robotics.maps.update(items)

    def delete(self, ids: SequenceNotStr[str]) -> int:
        try:
            self.client.robotics.maps.delete(ids)
        except CogniteAPIError as e:
            return len(e.successful)
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[Map]:
        return iter(self.client.robotics.maps)


def _fallback_to_one_by_one(
    api_call: Callable, items: Sequence | SequenceNotStr, return_cls: type[T_CogniteResourceList]
) -> T_CogniteResourceList:
    try:
        return api_call(items)
    except CogniteAPIError:
        return_items = return_cls([])
        if len(items) > 1:
            # The API does not give any information about which items were not found/failed.
            # so we need to apply them one by one to find out which ones failed.
            for item in items:
                with suppress(CogniteAPIError):
                    return_items.append(api_call(item))
        return return_items



from pathlib import Path

import typer

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables

from .commands.import_ import ImportTransformationCLI

import_app = typer.Typer(
    pretty_exceptions_short=False, pretty_exceptions_show_locals=False, pretty_exceptions_enable=False
)


@import_app.callback(invoke_without_command=True)
def import_main(ctx: typer.Context) -> None:
    """PREVIEW FEATURE Import resources into Cognite-Toolkit."""
    if ctx.invoked_subcommand is None:
        print("Use [bold yellow]cdf-tk import --help[/] for more information.")
    return None


@import_app.command("transformation-cli")
def transformation_cli(
    source: Path = typer.Argument(..., help="Path to the transformation CLI manifest directory or files."),
    destination: Path = typer.Argument(..., help="Path to the destination directory."),
    overwrite: bool = typer.Option(False, help="Overwrite destination if it already exists."),
    flatten: bool = typer.Option(False, help="Flatten the directory structure."),
    clean: bool = typer.Option(False, help="Remove the source directory after import."),
    verbose: bool = typer.Option(False, help="Turn on to get more verbose output when running the command"),
) -> None:
    """Import transformation CLI manifests into Cognite-Toolkit modules."""

    # We are lazy loading the client as we only need it if we need to look up dataset ids.
    # This is to ensure the command can be executed without a client if the user does not need to look up dataset ids.
    # (which is likely 99% of the time)
    def get_client() -> ToolkitClient:
        return EnvironmentVariables.create_from_environment().get_client()

    cmd = ImportTransformationCLI(print_warning=True, get_client=get_client)
    cmd.execute(source, destination, overwrite, flatten, clean, verbose=verbose)



from dataclasses import asdict, dataclass


@dataclass
class Dataclass:
    def to_dict(self, ignore_none=False):
        if ignore_none:
            return asdict(self, dict_factory=lambda x: {k: v for (k, v) in x if v})
        return asdict(self)


@dataclass
class Vec3f(Dataclass):
    x: float
    y: float
    z: float



from __future__ import annotations

import inspect
import itertools
import re
import shutil
from pathlib import Path
from typing import Any

import cognite.client.data_classes as data_classes
import cognite.client.data_classes.data_modeling as dm_data_classes
import yaml
from cognite.client.data_classes.data_modeling import ViewApply, ViewId
from cognite.client.data_classes.iam import Capability
from cognite.client.data_classes.transformations import OidcCredentials
from cognite.client.data_classes.transformations.common import TransformationDestination
from cognite.client.utils._text import to_camel_case

from cognite_toolkit._cdf_tk.loaders import ResourceLoader
from cognite_toolkit._cdf_tk.utils import safe_read
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml


# * in a line signifies that the line is a comment, for example, where the config has multiple options
def comment_optional(yaml_str: str) -> str:
    n = []
    for line in yaml_str.split("\n"):
        if "*" in line:
            n.append(f"#{line}".replace("*", "").replace("'", "").replace(" - :", ""))
        else:
            n.append(line.replace("*", ""))
    return "\n".join(n)


IGNORED_ANNOTATIONS = {
    "str",
    # "int",
    "float",
    "bool",
    "dict",
    "list",
    "None",
    "enum",
    "dict[str, str]",
    "list[str]",
    "Any",
    "Filter",
    "list[ExternalId]",
    "'allUserAccounts'",
    "list[ViewId]",
    "dict[str, ViewPropertyApply]",
}


def expand_acls() -> list[Any]:
    expanded = []
    for acl in itertools.chain(Capability.__subclasses__()):
        if acl.__name__ == "LegacyCapability" or acl.__name__ == "UnknownAcl" or acl.__name__ == "SeismicAcl":
            continue
        action_members = []
        for name, value in acl.Action.__dict__.items():
            if not name.startswith("_") and not callable(value):
                action_members.append(value.value)

        scope_members: dict[str, str | dict] = {}
        filtered_classes = {name: cls for name, cls in acl.Scope.__dict__.items() if not name.startswith("_")}
        for name, cls in filtered_classes.items():
            if cls._scope_name == "all":
                scope_members["all"] = {}
            else:
                if "ids" in cls.__annotations__:
                    scope_members[f"*{cls._scope_name}"] = {"*ids": "list[str]"}
                else:
                    scope_members[f"{cls._scope_name}"] = {}

        expanded.append(
            {
                acl._capability_name: {
                    "actions": action_members,
                    "scope": scope_members
                    if len(scope_members) == 1
                    else {"*one of": {"*" + k: v for k, v in scope_members.items()}},
                }
            }
        )
    return expanded


def expand_transformation_destinations() -> dict[str, Any]:
    options: dict[str, Any] = {}
    options["*type"] = []
    for name, member in inspect.getmembers(TransformationDestination):
        if isinstance(TransformationDestination.__dict__.get(name), staticmethod):
            signature = inspect.signature(member)
            if len(signature.parameters) == 0:
                options["*type"].append(f'*"{name}"')
            else:
                options[f"*{to_camel_case(name)}"] = {k: f"*{v.annotation}" for k, v in signature.parameters.items()}
    return {"*one of": options}


def get_parameters(cls: Any) -> Any:
    sign = inspect.signature(cls.__init__)

    filtered_params = [param for name, param in sign.parameters.items() if name != "self" and name != "cognite_client"]
    return sign.replace(parameters=filtered_params).parameters.values()


def unpack_annotations(str_annotations: str) -> Any:
    str_annotations = str_annotations.replace("Sequence[", "list[")
    str_annotations = re.sub(r"Literal\[(.*?)\]", lambda match: match.group(1), str_annotations)

    pattern = r"\|(?![^\[\]]*\])"
    annotations = [a.strip() for a in re.split(pattern, str_annotations)]

    if not any([a not in IGNORED_ANNOTATIONS for a in annotations]):
        return str_annotations

    cognite_classes = {name: obj for name, obj in inspect.getmembers(data_classes) if inspect.isclass(obj)}
    cognite_classes.update({name: obj for name, obj in inspect.getmembers(dm_data_classes) if inspect.isclass(obj)})

    resolved: dict[str, Any] = {}

    for annotation in annotations:
        if annotation.startswith("list["):
            return [unpack_annotations(re.sub(r"list\[(.*?)\]", lambda match: match.group(1), annotation))]
        else:
            cls = cognite_classes.get(annotation)
            if not cls:
                continue
            for param in get_parameters(cls):
                resolved[to_camel_case(param.name)] = unpack_annotations(str(param.annotation))

        return resolved

    return str_annotations


def expand_parameters(loader_cls: Any, cls: Any, optional: bool = False) -> dict[str, Any]:
    parameters = get_parameters(cls)
    params: dict[str, Any] = {}

    for param in parameters:
        # resource specifics:
        if cls.__name__ == "ExtractionPipelineWrite":
            if param.name == "raw_tables":
                params[to_camel_case(param.name)] = {
                    "*optional list[dict[str, str]]:": [
                        {"*db1": "*table1"},
                        {"*db2": "*table2"},
                        {"*db3": {}},
                    ]
                }
                continue

            if param.name == "contacts":
                params[param.name] = {
                    "*optional list[ExtractionPipelineContact] ": [
                        {"*name": "str", "*email": "*str", "*role": "str", "*sendNotification": "True | False"},
                        {"*name": "str", "*email": "*str", "*role": "str", "*sendNotification": "True | False"},
                    ]
                }
                continue

        elif cls.__name__ == "FunctionWrite":
            if param.name == "runtime":
                params[param.name] = "py38 | py39 | py310 | py311 | None"
                continue

            if param.name == "file_id":
                continue

        elif cls.__name__ == "FileMetadataWrite":
            if param.name == "labels":
                params[param.name] = {"*optional list[Label]": ["*label_1_ext_id", "*label_2_ext_id"]}
                continue

            if param.name == "geo_location":
                params[to_camel_case(param.name)] = {
                    "*optional GeoLocation": {
                        "*type": "'feature'",
                        "*geometry": "Point | MultiPoint | LineString | MultiLineString | Polygon | MultiPolygon",
                        "*properties": "dict | None",
                    }
                }
                continue

            if param.name == "asset_ids":
                params[to_camel_case(param.name)] = {"*optional list[int]": ["*asset_1_ext_id", "*asset_2_ext_id"]}
                continue

            if param.name == "security_categories":
                params[to_camel_case(param.name)] = {
                    "*optional list[SecurityCategory]": ["*sec_cat_1_ext_id", "*sec_cat_2_ext_id"]
                }
                continue

        elif cls.__name__ == "TransformationWrite":
            if param.name == "source_nonce" or param.name == "destination_nonce":
                continue

            if param.name == "conflict_mode":
                params[to_camel_case(param.name)] = {
                    "*optional": {"*one of": ['*"abort"', '*"*delete"', '*"update"', '*"upsert"']}
                }
                continue

            if param.name == "source_oidc_credentials":
                params["authentication"] = {
                    "*optional OidcCredentials": expand_parameters(loader_cls, OidcCredentials, optional=True)
                }
                continue

            if param.name == "destination_oidc_credentials":
                continue

            if param.name == "destination":
                params["destination"] = expand_transformation_destinations()
                continue

        elif cls.__name__ == "TimeSeriesWrite":
            if param.name == "instance_id":
                params[to_camel_case(param.name)] = [{"space": "str", "external_id": "str"}, "None"]
                continue

        elif cls.__name__ == "WorkflowDefinitionUpsert" or cls.__name__ == "WorkflowVersionUpsert":
            if param.name == "workflow_definition":
                params[to_camel_case(param.name)] = {
                    "description": "str | None",
                    "tasks": [
                        {
                            "externalId": "str",
                            "type": "function | transformation | cdf",
                            "parameters": {
                                "function": {
                                    "externalId": "str",
                                }
                            },
                        }
                    ],
                }
                continue

        elif cls.__name__ == "GroupWrite":
            if param.name == "capabilities":
                params[to_camel_case(param.name)] = expand_acls()
                continue

        elif cls.__name__ == "DataModelApply":
            if param.name == "views":
                params[to_camel_case(param.name)] = {
                    "*optional list[ViewId | ViewApply]": {
                        "*one list of": [expand_parameters(cls, ViewId, True), expand_parameters(cls, ViewApply, True)]
                    }
                }
                continue

        if str(param.annotation) == "dict[str, str] | None":
            params[to_camel_case(param.name)] = {
                "*optional dict[str,str]": {"*key1": "value1", "*key2": "value2"},
            }
            continue

        if param.name == "data_set_id":
            params["dataSetExternalId"] = param.annotation.replace("int", "str")
            continue

        params[to_camel_case(param.name)] = f"{'*' if optional else ''}{unpack_annotations(str(param.annotation))}"

    return params


def generate(target_path: Path) -> None:
    for loader_cls in [
        loader for loader in ResourceLoader.__subclasses__() if loader.__name__ not in ["ResourceContainerLoader"]
    ]:
        write_cls = loader_cls.resource_write_cls  # type: ignore

        ref = expand_parameters(loader_cls, write_cls)

        final_yaml = comment_optional(yaml.dump(ref, sort_keys=False))

        # special case for groups: want caps at the end
        if "capabilities" in ref:
            caps_yaml = comment_optional(yaml.dump({"capabilities": ref.pop("capabilities")}, sort_keys=False))
            main_yaml = comment_optional(yaml.dump(ref, sort_keys=False))
            final_yaml = f"{main_yaml}{caps_yaml}"

        folder = target_path / loader_cls.folder_name
        file_name = folder / Path(f"reference.{loader_cls.kind}.yaml")
        Path.mkdir(folder, exist_ok=True)
        Path.write_text(file_name, final_yaml)

        print(f"Wrote {file_name}")

        print(f"Wrote {file_name}")


def validate(target_path: Path) -> None:
    total_warnings = 0
    for loader_cls in [
        loader for loader in ResourceLoader.__subclasses__() if loader.__name__ not in ["ResourceContainerLoader"]
    ]:
        folder = target_path / loader_cls.folder_name
        file_name = folder / Path(f"reference.{loader_cls.kind}.yaml")
        warnings = validate_resource_yaml(
            data=yaml.safe_load(safe_read(file_name)),
            spec=loader_cls.get_write_cls_parameter_spec(),
            source_file=file_name,
        )

        total_warnings += len(warnings)
        if len(warnings) > 0:
            print(f"Warnings for {file_name}")
            for warning in warnings:
                print(warning)
            break

    print(f"Total warnings: {total_warnings}")


if __name__ == "__main__":
    target_path = Path("cognite_toolkit/_cdf_tk/prototypes/_packages/reference/references")
    if Path.exists(target_path):
        shutil.rmtree(target_path)
    Path.mkdir(target_path)

    generate(target_path)
    validate(target_path)
    manifest_file = target_path.parent / Path("manifest.yaml")
    manifest_content = {
        "title": "References: reference yaml files for the supported resources",
        "modules": {
            "references": {
                "title": "Reference files",
            },
        },
    }
    manifest_file.write_text(yaml.dump(manifest_content, sort_keys=False))



from pathlib import Path
from typing import Annotated, Any, Optional

import typer
from rich import print

from cognite_toolkit._cdf_tk.commands import PopulateCommand
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables


class PopulateApp(typer.Typer):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.callback(invoke_without_command=True)(self.main)
        self.command("view")(self.populate_view)

    def main(self, ctx: typer.Context) -> None:
        """Commands populate functionality"""
        if ctx.invoked_subcommand is None:
            print("Use [bold yellow]cdf populate --help[/] for more information.")

    def populate_view(
        self,
        ctx: typer.Context,
        view_id: Annotated[
            Optional[list[str]],
            typer.Argument(
                help="View ID to populate. Format space external ID version, for example 'cdf_cdm CogniteAsset v1'"
                " If not provided, interactive mode will be used.",
            ),
        ] = None,
        table: Annotated[
            Optional[Path],
            typer.Option(
                "--table",
                "-t",
                help="Path to the file containing the data to populate the view with. This is required unless interactive"
                " mode is used",
            ),
        ] = None,
        instance_space: Annotated[
            Optional[str],
            typer.Option(
                "--space",
                "-s",
                help="The space to write the nodes to. This is required unless interactive mode is used",
            ),
        ] = None,
        external_id_column: Annotated[
            Optional[str],
            typer.Option(
                "--external-id-column",
                "-e",
                help="The name of the column in the table that contains the external IDs of the nodes. The column"
                " must be present in the table and must contain unique values. This is required unless interactive"
                " mode is used",
            ),
        ] = None,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will populate a given view with data from a local table."""
        cmd = PopulateCommand()
        client = EnvironmentVariables.create_from_environment().get_client()
        cmd.run(
            lambda: cmd.view(
                client,
                view_id,
                table,
                instance_space,
                external_id_column,
                verbose,
            )
        )



from __future__ import annotations

import uuid
from collections.abc import Hashable
from graphlib import CycleError, TopologicalSorter
from typing import cast

import questionary
from cognite.client.data_classes import AggregateResultItem, DataSetUpdate, filters
from cognite.client.data_classes.data_modeling import NodeId
from cognite.client.exceptions import CogniteAPIError
from rich import print
from rich.console import Console
from rich.panel import Panel
from rich.status import Status

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.data_classes import DeployResults, ResourceDeployResult
from cognite_toolkit._cdf_tk.exceptions import (
    CDFAPIError,
    ToolkitMissingResourceError,
    ToolkitRequiredValueError,
    ToolkitValueError,
)
from cognite_toolkit._cdf_tk.loaders import (
    RESOURCE_LOADER_LIST,
    AssetLoader,
    CogniteFileLoader,
    DataSetsLoader,
    FunctionLoader,
    GraphQLLoader,
    GroupAllScopedLoader,
    GroupLoader,
    GroupResourceScopedLoader,
    HostedExtractorDestinationLoader,
    LocationFilterLoader,
    NodeLoader,
    ResourceLoader,
    SpaceLoader,
    StreamlitLoader,
    TransformationLoader,
    ViewLoader,
)
from cognite_toolkit._cdf_tk.tk_warnings import HighSeverityWarning, MediumSeverityWarning
from cognite_toolkit._cdf_tk.utils import humanize_collection

from ._base import ToolkitCommand


class PurgeCommand(ToolkitCommand):
    def space(
        self,
        client: ToolkitClient,
        space: str | None = None,
        include_space: bool = False,
        dry_run: bool = False,
        auto_yes: bool = False,
        verbose: bool = False,
    ) -> None:
        """Purge a space and all its content"""
        selected_space = self._get_selected_space(space, client)
        if space is None:
            # Interactive mode
            include_space = questionary.confirm("Do you also want to delete the space itself?", default=False).ask()
            dry_run = questionary.confirm("Dry run?", default=True).ask()
        if not dry_run:
            self._print_panel("space", selected_space)
            if not auto_yes:
                confirm = questionary.confirm(
                    f"Are you really sure you want to purge the {selected_space!r} space?", default=False
                ).ask()
                if not confirm:
                    return

        loaders = self._get_dependencies(
            SpaceLoader,
            exclude={
                GraphQLLoader,
                GroupResourceScopedLoader,
                LocationFilterLoader,
                TransformationLoader,
                CogniteFileLoader,
            },
        )
        is_purged = self._purge(client, loaders, selected_space, dry_run=dry_run, verbose=verbose)
        if include_space and is_purged:
            space_loader = SpaceLoader.create_loader(client)
            if dry_run:
                print(f"Would delete space {selected_space}")
            else:
                try:
                    space_loader.delete([selected_space])
                    print(f"Space {selected_space} deleted")
                except CogniteAPIError as e:
                    self.warn(HighSeverityWarning(f"Failed to delete space {selected_space!r}: {e}"))
        elif include_space:
            self.warn(HighSeverityWarning(f"Space {selected_space!r} was not deleted due to errors during the purge"))

        if not dry_run and is_purged:
            print(f"Purge space {selected_space!r} completed.")
        elif not dry_run:
            print(f"Purge space {selected_space!r} partly completed. See warnings for details.")

    @staticmethod
    def _get_dependencies(
        loader_cls: type[ResourceLoader], exclude: set[type[ResourceLoader]] | None = None
    ) -> dict[type[ResourceLoader], frozenset[type[ResourceLoader]]]:
        return {
            dep_cls: dep_cls.dependencies
            for dep_cls in RESOURCE_LOADER_LIST
            if loader_cls in dep_cls.dependencies and (exclude is None or dep_cls not in exclude)
        }

    @staticmethod
    def _get_selected_space(space: str | None, client: ToolkitClient) -> str:
        if space is None:
            spaces = client.data_modeling.spaces.list(limit=-1, include_global=False)
            selected_space = questionary.select(
                "Which space do you want to purge"
                " (including all data models, views, containers, nodes and edges within that space)?",
                sorted([space.space for space in spaces]),
            ).ask()
        else:
            retrieved = client.data_modeling.spaces.retrieve(space)
            if retrieved is None:
                raise ToolkitMissingResourceError(f"Space {space} does not exist")
            selected_space = space

        if selected_space is None:
            raise ToolkitValueError("No space selected")
        return selected_space

    def dataset(
        self,
        client: ToolkitClient,
        external_id: str | None = None,
        include_dataset: bool = False,
        dry_run: bool = False,
        auto_yes: bool = False,
        verbose: bool = False,
    ) -> None:
        """Purge a dataset and all its content"""
        selected_dataset = self._get_selected_dataset(external_id, client)
        if external_id is None:
            # Interactive mode
            include_dataset = questionary.confirm(
                "Do you want to archive the dataset itself after the purge?", default=False
            ).ask()
            dry_run = questionary.confirm("Dry run?", default=True).ask()
        if not dry_run:
            self._print_panel("dataset", selected_dataset)
            if not auto_yes:
                confirm = questionary.confirm(
                    f"Are you really sure you want to purge the {selected_dataset!r} dataset?", default=False
                ).ask()
                if not confirm:
                    return

        loaders = self._get_dependencies(
            DataSetsLoader,
            exclude={
                GroupLoader,
                GroupResourceScopedLoader,
                GroupAllScopedLoader,
                StreamlitLoader,
                HostedExtractorDestinationLoader,
                FunctionLoader,
                LocationFilterLoader,
            },
        )
        is_purged = self._purge(client, loaders, selected_data_set=selected_dataset, dry_run=dry_run, verbose=verbose)
        if include_dataset and is_purged:
            if dry_run:
                print(f"Would have archived {selected_dataset}")
            else:
                archived = (
                    DataSetUpdate(external_id=selected_dataset)
                    .external_id.set(str(uuid.uuid4()))
                    .metadata.add({"archived": "true"})
                    .write_protected.set(True)
                )
                client.data_sets.update(archived)
                print(f"DataSet {selected_dataset} archived")
        elif include_dataset:
            self.warn(
                HighSeverityWarning(f"DataSet {selected_dataset} was not archived due to errors during the purge")
            )

        if not dry_run and is_purged:
            print(f"Purged dataset {selected_dataset!r} completed")
        elif not dry_run:
            print(f"Purged dataset {selected_dataset!r} partly completed. See warnings for details.")

    def _print_panel(self, resource_type: str, resource: str) -> None:
        print(
            Panel(
                f"[red]WARNING:[/red] This operation [bold]cannot be undone[/bold]! "
                f"Resources in {resource!r} are permanently deleted",
                style="bold",
                title=f"Purge {resource_type}",
                title_align="left",
                border_style="red",
                expand=False,
            )
        )

    @staticmethod
    def _get_selected_dataset(external_id: str | None, client: ToolkitClient) -> str:
        if external_id is None:
            datasets = client.data_sets.list(limit=-1)
            selected_dataset: str = questionary.select(
                "Which space are you going to purge (delete all resources in dataset)?",
                sorted([dataset.external_id for dataset in datasets if dataset.external_id]),
            ).ask()
        else:
            retrieved = client.data_sets.retrieve(external_id=external_id)
            if retrieved is None:
                raise ToolkitMissingResourceError(f"DataSet {external_id!r} does not exist")
            selected_dataset = external_id

        if selected_dataset is None:
            raise ToolkitValueError("No space selected")
        return selected_dataset

    def _purge(
        self,
        client: ToolkitClient,
        loaders: dict[type[ResourceLoader], frozenset[type[ResourceLoader]]],
        selected_space: str | None = None,
        selected_data_set: str | None = None,
        dry_run: bool = False,
        verbose: bool = False,
        batch_size: int = 1000,
    ) -> bool:
        is_purged = True
        results = DeployResults([], "purge", dry_run=dry_run)
        loader_cls: type[ResourceLoader]
        has_purged_views = False
        with Console().status("...", spinner="aesthetic", speed=0.4) as status:
            for loader_cls in reversed(list(TopologicalSorter(loaders).static_order())):
                if loader_cls not in loaders:
                    # Dependency that is included
                    continue
                loader = loader_cls.create_loader(client, console=status.console)
                status_prefix = "Would have deleted" if dry_run else "Deleted"
                if isinstance(loader, ViewLoader) and not dry_run:
                    status_prefix = "Expected deleted"  # Views are not always deleted immediately
                    has_purged_views = True

                if not dry_run and isinstance(loader, NodeLoader):
                    # Special handling of nodes as node type must be deleted after regular nodes
                    # In dry-run mode, we are not deleting the nodes, so we can skip this.
                    warnings_before = len(self.warning_list)
                    deleted_nodes = self._purge_nodes(loader, status, selected_space, verbose)
                    results[loader.display_name] = ResourceDeployResult(
                        name=loader.display_name,
                        deleted=deleted_nodes,
                        total=deleted_nodes,
                    )
                    if len(self.warning_list) > warnings_before and any(
                        isinstance(warn, HighSeverityWarning) for warn in self.warning_list[warnings_before:]
                    ):
                        is_purged = False
                    continue
                elif not dry_run and isinstance(loader, AssetLoader):
                    # Special handling of assets as we must ensure all children are deleted before the parent.
                    # In dry-run mode, we are not deleting the assets, so we can skip this.
                    deleted_assets = self._purge_assets(loader, status, selected_data_set)
                    results[loader.display_name] = ResourceDeployResult(
                        loader.display_name, deleted=deleted_assets, total=deleted_assets
                    )
                    continue

                # Child loaders are, for example, WorkflowTriggerLoader, WorkflowVersionLoader for WorkflowLoader
                # These must delete all resources that are connected to the resource that the loader is deleting
                # Exclude loaders that we are already iterating over
                child_loader_classes = self._get_dependencies(loader_cls, exclude=set(loaders))
                child_loaders = [
                    child_loader.create_loader(client)
                    for child_loader in reversed(list(TopologicalSorter(child_loader_classes).static_order()))
                    # Necessary as the topological sort includes dependencies that are not in the loaders
                    if child_loader in child_loader_classes
                ]
                count = 0
                status.update(f"{status_prefix} {count:,} {loader.display_name}...")
                batch_ids: list[Hashable] = []
                for resource in loader.iterate(data_set_external_id=selected_data_set, space=selected_space):
                    try:
                        batch_ids.append(loader.get_id(resource))
                    except (ToolkitRequiredValueError, KeyError) as e:
                        try:
                            batch_ids.append(loader.get_internal_id(resource))
                        except (AttributeError, NotImplementedError):
                            self.warn(
                                HighSeverityWarning(
                                    f"Cannot delete {type(resource).__name__}. Failed to obtain ID: {e}"
                                ),
                                console=status.console,
                            )
                            is_purged = False
                            continue

                    if len(batch_ids) >= batch_size:
                        child_deletion = self._delete_children(
                            batch_ids, child_loaders, dry_run, status.console, verbose
                        )
                        batch_delete, batch_size = self._delete_batch(
                            batch_ids, dry_run, loader, batch_size, status.console, verbose
                        )
                        count += batch_delete
                        status.update(f"{status_prefix} {count:,} {loader.display_name}...")
                        batch_ids = []
                        # The DeployResults is overloaded such that the below accumulates the counts
                        for name, child_count in child_deletion.items():
                            results[name] = ResourceDeployResult(name, deleted=child_count, total=child_count)

                if batch_ids:
                    child_deletion = self._delete_children(batch_ids, child_loaders, dry_run, status.console, verbose)
                    batch_delete, batch_size = self._delete_batch(
                        batch_ids, dry_run, loader, batch_size, status.console, verbose
                    )
                    count += batch_delete
                    status.update(f"{status_prefix} {count:,} {loader.display_name}...")
                    for name, child_count in child_deletion.items():
                        results[name] = ResourceDeployResult(name, deleted=child_count, total=child_count)
                if count > 0:
                    status.console.print(f"{status_prefix} {count:,} {loader.display_name}.")
                results[loader.display_name] = ResourceDeployResult(
                    name=loader.display_name,
                    deleted=count,
                    total=count,
                )
        print(results.counts_table(exclude_columns={"Created", "Changed", "Untouched", "Total"}))
        if has_purged_views:
            print("You might need to run the purge command multiple times to delete all views.")
        return is_purged

    def _delete_batch(
        self,
        batch_ids: list[Hashable],
        dry_run: bool,
        loader: ResourceLoader,
        batch_size: int,
        console: Console,
        verbose: bool,
    ) -> tuple[int, int]:
        if dry_run:
            deleted = len(batch_ids)
        else:
            try:
                deleted = loader.delete(batch_ids)
            except CogniteAPIError as delete_error:
                if (
                    delete_error.code == 408
                    and "timed out" in delete_error.message.casefold()
                    and batch_size > 1
                    and (len(batch_ids) > 1)
                ):
                    self.warn(
                        MediumSeverityWarning(
                            f"Timed out deleting {loader.display_name}. Trying again with a smaller batch size."
                        ),
                        include_timestamp=True,
                        console=console,
                    )
                    new_batch_size = len(batch_ids) // 2
                    first = batch_ids[:new_batch_size]
                    second = batch_ids[new_batch_size:]
                    first_deleted, first_batch_size = self._delete_batch(
                        first, dry_run, loader, new_batch_size, console, verbose
                    )
                    second_deleted, second_batch_size = self._delete_batch(
                        second, dry_run, loader, new_batch_size, console, verbose
                    )
                    return first_deleted + second_deleted, min(first_batch_size, second_batch_size)
                else:
                    raise delete_error

        if verbose:
            prefix = "Would delete" if dry_run else "Finished purging"
            console.print(f"{prefix} {deleted:,} {loader.display_name}")
        return deleted, batch_size

    @staticmethod
    def _delete_children(
        parent_ids: list[Hashable], child_loaders: list[ResourceLoader], dry_run: bool, console: Console, verbose: bool
    ) -> dict[str, int]:
        child_deletion: dict[str, int] = {}
        for child_loader in child_loaders:
            child_ids = set()
            for child in child_loader.iterate(parent_ids=parent_ids):
                child_ids.add(child_loader.get_id(child))
            count = 0
            if child_ids:
                if dry_run:
                    count = len(child_ids)
                else:
                    count = child_loader.delete(list(child_ids))

                if verbose:
                    prefix = "Would delete" if dry_run else "Deleted"
                    console.print(f"{prefix} {count:,} {child_loader.display_name}")
            child_deletion[child_loader.display_name] = count
        return child_deletion

    def _purge_nodes(
        self,
        loader: NodeLoader,
        status: Status,
        selected_space: str | None = None,
        verbose: bool = False,
        batch_size: int = 1000,
    ) -> int:
        """Special handling of nodes as we must ensure all node types are deleted last."""
        # First find all Node Types
        node_types: dict[NodeId, set[NodeId]] = {}
        total_count = 0
        for node in loader.iterate(space=selected_space):
            if node.type and (selected_space is None or node.space == selected_space):
                node_id = NodeId(node.type.space, node.type.external_id)
                node_types[node_id] = set()
            total_count += 1
            status.update(f"Looking up node.type {total_count:,}...")

        count = 0
        batch_ids: list[NodeId] = []
        for node in loader.iterate(space=selected_space):
            node_id = node.as_id()
            if node_id in node_types:
                # Skip if it is a node type
                continue
            batch_ids.append(node_id)
            if len(batch_ids) >= batch_size:
                deleted, batch_size = self._delete_node_batch(batch_ids, loader, batch_size, status.console, verbose)
                count += deleted
                status.update(f"Deleted {count:,}/{total_count} {loader.display_name}...")
                batch_ids = []

        if batch_ids:
            deleted, batch_size = self._delete_node_batch(batch_ids, loader, batch_size, status.console, verbose)
            count += deleted

        # Finally delete all node types, first do a lookup and topological sort to ensure deleting in the right order.
        # Note this is an edge case, and is a result of some strange usage. But we need to handle it.
        for node_type in loader.retrieve(list(node_types.keys())):
            if node_type.type:
                required = NodeId(node_type.type.space, node_type.type.external_id)
                if required in node_types:
                    node_types[required].add(node_type.as_id())

        try:
            node_type_ids = list(TopologicalSorter(node_types).static_order())
        except CycleError as e:
            self.warn(
                HighSeverityWarning(f"Failed to delete node-types: Cycle detected. {e.args}"), console=status.console
            )
            return count

        deleted, batch_size = self._delete_node_batch(node_type_ids, loader, batch_size, status.console, verbose)
        count += deleted
        if count > 0:
            status.console.print(f"Finished purging {loader.display_name}.")
        return count

    def _delete_node_batch(
        self, batch_ids: list[NodeId], loader: NodeLoader, batch_size: int, console: Console, verbose: bool
    ) -> tuple[int, int]:
        try:
            deleted = loader.delete(batch_ids)
        except CogniteAPIError as delete_error:
            if (
                delete_error.code == 400
                and "Attempted to delete a node which is used as a type" in delete_error.message
            ):
                # Fallback to delete one by one
                deleted = 0
                for node_id in batch_ids:
                    try:
                        loader.delete([node_id])
                        deleted += 1
                    except CogniteAPIError:
                        is_type = filters.Equals(["node", "type"], node_id.dump(include_instance_type=False))
                        instance_spaces = {node.space for node in loader.client.data_modeling.instances(filter=is_type)}
                        if instance_spaces:
                            suffix = (
                                "It is used as a node type in the following spaces, "
                                f"which must be purged first: {humanize_collection(instance_spaces)}"
                            )
                        else:
                            suffix = "It is used as a node type, but failed to find the spaces where it is used."
                        self.warn(HighSeverityWarning(f"Failed to delete {node_id!r}. {suffix}"), console=console)
            elif (
                delete_error.code == 408
                and "timed out" in delete_error.message.casefold()
                and batch_size > 1
                and (len(batch_ids) > 1)
            ):
                self.warn(
                    MediumSeverityWarning(
                        f"Timed out deleting {loader.display_name}. Trying again with a smaller batch size."
                    ),
                    include_timestamp=True,
                    console=console,
                )
                new_batch_size = len(batch_ids) // 2
                first = batch_ids[:new_batch_size]
                second = batch_ids[new_batch_size:]
                first_deleted, first_batch_size = self._delete_node_batch(
                    first, loader, new_batch_size, console, verbose
                )
                second_deleted, second_batch_size = self._delete_node_batch(
                    second, loader, new_batch_size, console, verbose
                )
                return first_deleted + second_deleted, min(first_batch_size, second_batch_size)
            else:
                raise delete_error

        if verbose:
            console.print(f"Deleted {deleted:,} {loader.display_name}")
        return deleted, batch_size

    def _purge_assets(
        self,
        loader: AssetLoader,
        status: Status,
        selected_data_set: str | None = None,
        batch_size: int = 1000,
    ) -> int:
        # Using sets to avoid duplicates
        children_ids: set[int] = set()
        parent_ids: set[int] = set()
        is_first = True
        last_failed = False
        count = level = depth = last_parent_count = 0
        total_asset_count: int | None = None
        # Iterate through the asset hierarchy once per depth level. This is to delete all children before the parent.
        while is_first or level < depth:
            for asset in loader.iterate(data_set_external_id=selected_data_set):
                aggregates = cast(AggregateResultItem, asset.aggregates)
                if is_first and aggregates.depth is not None:
                    depth = max(depth, aggregates.depth)

                if aggregates.child_count == 0:
                    children_ids.add(asset.id)
                else:
                    parent_ids.add(asset.id)

                if len(children_ids) >= batch_size:
                    count += loader.delete(list(children_ids))
                    if total_asset_count:
                        status.update(f"Deleted {count:,}/{total_asset_count:,} {loader.display_name}...")
                    else:
                        status.update(f"Deleted {count:,} {loader.display_name}...")
                    children_ids = set()

            if is_first:
                total_asset_count = count + len(parent_ids) + len(children_ids)
            if children_ids:
                count += loader.delete(list(children_ids))
                status.update(f"Deleted {count:,}/{total_asset_count:,} {loader.display_name}...")
                children_ids = set()

            if len(parent_ids) == last_parent_count and last_failed:
                try:
                    # Just try to delete them all at once
                    count += loader.delete(list(parent_ids))
                except CogniteAPIError as e:
                    raise CDFAPIError(
                        f"Failed to delete {len(parent_ids)} assets. This could be due to a parent-child cycle or an "
                        "eventual consistency issue. Wait a few seconds and try again. An alternative is to use the "
                        "Python-SDK to delete the asset hierarchy "
                        "`client.assets.delete(external_id='my_root_asset', recursive=True)`"
                    ) from e
                else:
                    status.update(f"Deleted {count:,}/{total_asset_count:,} {loader.display_name}...")
                    break
            elif len(parent_ids) == last_parent_count:
                last_failed = True
            else:
                last_failed = False
            level += 1
            last_parent_count = len(parent_ids)
            parent_ids.clear()
            is_first = False
        status.console.print(f"Finished purging {loader.display_name}.")
        return count



import json
import sys
import traceback
from collections import defaultdict
from collections.abc import Iterable, Sequence
from typing import ClassVar, Literal, TypeVar, cast

from cognite.client.config import global_config
from cognite.client.exceptions import CogniteAPIError

# Do not warn the user about feature previews from the Cognite-SDK we use in Toolkit
# ruff: noqa: E402
global_config.disable_pypi_version_check = True
global_config.silence_feature_preview_warnings = True
import yaml
from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes import ExtractionPipelineRunWrite, RowWrite
from cognite.client.data_classes.data_modeling.cdm.v1 import CogniteDiagramAnnotation, CogniteDiagramAnnotationApply
from pydantic import BaseModel, model_validator
from pydantic.alias_generators import to_camel

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


FUNCTION_ID = "connection_writer"
EXTRACTION_PIPELINE_EXTERNAL_ID = "ctx_files_direct_relation_write"
EXTERNAL_ID_LIMIT = 256
EXTRACTION_RUN_MESSAGE_LIMIT = 1000


def handle(data: dict, client: CogniteClient) -> dict:
    try:
        connection_count = execute(data, client)
    except Exception as e:
        tb = traceback.extract_tb(e.__traceback__)
        last_entry_this_file = next((entry for entry in reversed(tb) if entry.filename == __file__), None)
        suffix = ""
        if last_entry_this_file:
            suffix = f" in function {last_entry_this_file.name} on line {last_entry_this_file.lineno}: {last_entry_this_file.line}"

        status: Literal["failure", "success"] = "failure"
        # Truncate the error message to 1000 characters the maximum allowed by the API
        prefix = f"ERROR {FUNCTION_ID}: "
        error_msg = f'"{e!s}"'
        message = prefix + error_msg + suffix
        if len(message) >= EXTRACTION_RUN_MESSAGE_LIMIT:
            error_msg = error_msg[: EXTRACTION_RUN_MESSAGE_LIMIT - len(prefix) - len(suffix) - 3 - 1]
            message = prefix + error_msg + '..."' + suffix
    else:
        status = "success"
        message = f"{FUNCTION_ID} executed successfully. Created {connection_count} connections"

    client.extraction_pipelines.runs.create(
        ExtractionPipelineRunWrite(extpipe_external_id=EXTRACTION_PIPELINE_EXTERNAL_ID, status=status, message=message)
    )
    # Need to run at least daily or the sync endpoint will forget the cursors
    # (max time is 3 days).
    return {"status": status, "message": message}


################# Data Classes #################


# Logger using print
class CogniteFunctionLogger:
    def __init__(self, log_level: Literal["DEBUG", "INFO", "WARNING", "ERROR"] = "INFO"):
        self.log_level = log_level.upper()

    def _print(self, prefix: str, message: str) -> None:
        if "\n" not in message:
            print(f"{prefix} {message}")
            return
        lines = message.split("\n")
        print(f"{prefix} {lines[0]}")
        prefix_len = len(prefix)
        for line in lines[1:]:
            print(f"{' ' * prefix_len} {line}")

    def debug(self, message: str) -> None:
        if self.log_level == "DEBUG":
            self._print("[DEBUG]", message)

    def info(self, message: str) -> None:
        if self.log_level in ("DEBUG", "INFO"):
            self._print("[INFO]", message)

    def warning(self, message: str) -> None:
        if self.log_level in ("DEBUG", "INFO", "WARNING"):
            self._print("[WARNING]", message)

    def error(self, message: str) -> None:
        self._print("[ERROR]", message)


class ViewProperty(BaseModel, alias_generator=to_camel):
    space: str
    external_id: str
    version: str
    direct_relation_property: str | None = None

    def as_view_id(self) -> dm.ViewId:
        return dm.ViewId(space=self.space, external_id=self.external_id, version=self.version)


class DirectRelationMapping(BaseModel, alias_generator=to_camel):
    start_node_view: ViewProperty
    end_node_view: ViewProperty

    @model_validator(mode="after")
    def direct_relation_is_set(self) -> Self:
        if (
            sum(
                1
                for prop in (self.start_node_view.direct_relation_property, self.end_node_view.direct_relation_property)
                if prop is not None
            )
            != 1
        ):
            raise ValueError("You must set 'directRelationProperty' for at either of 'startNode' or 'endNode'")
        return self


class ConfigData(BaseModel, alias_generator=to_camel):
    annotation_space: str
    direct_relation_mappings: list[DirectRelationMapping]


class ConfigState(BaseModel, alias_generator=to_camel):
    raw_database: str
    raw_table: str


class Config(BaseModel, alias_generator=to_camel):
    data: ConfigData
    state: ConfigState


class State(BaseModel):
    key: ClassVar[str] = FUNCTION_ID
    last_cursor: str | None = None

    @classmethod
    def from_cdf(cls, client: CogniteClient, state: ConfigState) -> "State":
        row = client.raw.rows.retrieve(db_name=state.raw_database, table_name=state.raw_table, key=cls.key)
        if row is None:
            return cls()
        return cls.model_validate(row.columns)

    def to_cdf(self, client: CogniteClient, state: ConfigState) -> None:
        client.raw.rows.insert(
            db_name=state.raw_database,
            table_name=state.raw_table,
            row=self._as_row(),
        )

    def _as_row(self) -> RowWrite:
        return RowWrite(
            key=self.key,
            columns=self.model_dump(),
        )


################# Functions #################


def execute(data: dict, client: CogniteClient) -> int:
    logger = CogniteFunctionLogger(data.get("logLevel", "INFO"))  # type: ignore[arg-type]
    logger.debug("Starting connection write")
    config = load_config(client, logger)
    logger.debug("Loaded config successfully")

    state = State.from_cdf(client, config.state)
    connection_count = 0
    for annotation_list in iterate_new_approved_annotations(state, client, config.data.annotation_space, logger):
        annotation_by_source_by_node = to_direct_relations_by_source_by_node(
            annotation_list, config.data.direct_relation_mappings, logger
        )
        connections = write_connections(annotation_by_source_by_node, client, logger)
        connection_count += connections

    state.to_cdf(client, config.state)
    logger.info(f"Created {connection_count} connections")
    return connection_count


T = TypeVar("T")


def chunker(items: Sequence[T], chunk_size: int) -> Iterable[Sequence[T]]:
    for i in range(0, len(items), chunk_size):
        yield items[i : i + chunk_size]


def iterate_new_approved_annotations(
    state: State, client: CogniteClient, annotation_space: str, logger: CogniteFunctionLogger, chunk_size: int = 1000
) -> Iterable[list[CogniteDiagramAnnotation]]:
    query = create_query(state.last_cursor, annotation_space)
    try:
        result = client.data_modeling.instances.sync(query)
    except CogniteAPIError as e:
        if e.code == 400 and "Cursor has expired" in e.message:
            logger.warning("Cursor has expired, starting from the beginning")
            state.last_cursor = None
            query = create_query(state.last_cursor, annotation_space)
            result = client.data_modeling.instances.sync(query)
        else:
            raise
    edges = result["annotations"]
    logger.debug(f"Retrieved {len(edges)} new approved annotations")
    state.last_cursor = edges.cursor
    for edge_list in chunker(list(edges), chunk_size):
        yield [CogniteDiagramAnnotation._load(edge.dump()) for edge in edge_list]


def write_connections(
    annotation_by_source_by_property_by_view: dict[
        dm.ViewId, dict[tuple[dm.NodeId, str], list[dm.DirectRelationReference]]
    ],
    client: CogniteClient,
    logger: CogniteFunctionLogger,
) -> int:
    connection_count = 0
    updated_nodes: list[dm.NodeApply] = []
    for view_id, annotation_by_source_by_property in annotation_by_source_by_property_by_view.items():
        node_ids = [node_id for node_id, _ in annotation_by_source_by_property.keys()]
        existing_node_list = client.data_modeling.instances.retrieve(node_ids, sources=[view_id]).nodes
        existing_node_by_id = {node.as_id(): node.as_write() for node in existing_node_list}

        for (node_id, direct_relation_property), direct_relation_ids in annotation_by_source_by_property.items():
            existing_node = existing_node_by_id.get(node_id)
            if existing_node is None:
                logger.warning(f"Node {node_id} not found in view {view_id}")
                continue
            for entity_source in existing_node.sources:
                if entity_source.source == view_id:
                    existing_connections = cast(list[dict], entity_source.properties.get(direct_relation_property, []))
                    before = len(existing_connections)
                    all_connections = {
                        dm.DirectRelationReference.load(connection) for connection in existing_connections
                    } | set(direct_relation_ids)
                    after = len(all_connections)
                    entity_source.properties[direct_relation_property] = [  # type: ignore[index]
                        connection.dump() for connection in all_connections
                    ]
                    connection_count += after - before
                    break
            updated_nodes.append(existing_node)

    updated = client.data_modeling.instances.apply(updated_nodes)
    logger.debug(f"Updated {len(updated.nodes)} nodes")
    return connection_count


def to_direct_relations_by_source_by_node(
    annotations: list[CogniteDiagramAnnotation], mappings: list[DirectRelationMapping], logger: CogniteFunctionLogger
) -> dict[dm.ViewId, dict[tuple[dm.NodeId, str], list[dm.DirectRelationReference]]]:
    mapping_by_entity_source: dict[tuple[dm.ViewId, dm.ViewId], DirectRelationMapping] = {
        (mapping.start_node_view.as_view_id(), mapping.end_node_view.as_view_id()): mapping for mapping in mappings
    }
    annotation_by_source_by_node: dict[dm.ViewId, dict[tuple[dm.NodeId, str], list[dm.DirectRelationReference]]] = (
        defaultdict(lambda: defaultdict(list))
    )
    for annotation in annotations:
        try:
            source_context = json.loads(annotation.source_context)
        except json.JSONDecodeError:
            logger.error(f"Could not parse source context for annotation {annotation.external_id}")
            continue
        try:
            start_view = dm.ViewId.load(source_context["start"])
            end_view = dm.ViewId.load(source_context["end"])
        except KeyError:
            logger.error(f"Missing start or end in source context for annotation {annotation.external_id}")
            continue
        mapping = mapping_by_entity_source.get((start_view, end_view))
        if mapping is None:
            logger.warning(
                f"No mapping found for entity source {(start_view, end_view)} for annotation {annotation.external_id}"
            )
            continue
        if mapping.start_node_view.direct_relation_property is not None:
            update_node = annotation.start_node
            direct_relation_property = mapping.start_node_view.direct_relation_property
            other_side = annotation.end_node
            view_id = mapping.start_node_view.as_view_id()
        elif mapping.end_node_view.direct_relation_property is not None:
            update_node = annotation.end_node
            direct_relation_property = mapping.end_node_view.direct_relation_property
            other_side = annotation.start_node
            view_id = mapping.end_node_view.as_view_id()
        else:
            raise ValueError(
                f"Neither file source nor entity source has a direct relation property for annotation {annotation.external_id}"
            )
        node = dm.NodeId(update_node.space, update_node.external_id)
        annotation_by_source_by_node[view_id][(node, direct_relation_property)].append(
            dm.DirectRelationReference(other_side.space, other_side.external_id)
        )
    return annotation_by_source_by_node


def create_query(last_cursor: str | None, annotation_space: str) -> dm.query.Query:
    view_id = CogniteDiagramAnnotationApply.get_source()
    is_annotation = dm.filters.And(
        dm.filters.Equals(["edge", "space"], annotation_space),
        dm.filters.HasData(views=[view_id]),
        dm.filters.Equals(view_id.as_property_ref("status"), "Approved"),
    )
    return dm.query.Query(
        with_={
            "annotations": dm.query.EdgeResultSetExpression(
                from_=None,
                filter=is_annotation,
                limit=1000,
            )
        },
        select={
            "annotations": dm.query.Select(
                [
                    dm.query.SourceSelector(
                        source=CogniteDiagramAnnotationApply.get_source(), properties=["sourceContext"]
                    )
                ]
            )
        },
        cursors={"annotations": last_cursor},
    )


def load_config(client: CogniteClient, logger: CogniteFunctionLogger) -> Config:
    raw_config = client.extraction_pipelines.config.retrieve(EXTRACTION_PIPELINE_EXTERNAL_ID)
    if raw_config.config is None:
        raise ValueError(f"Config for extraction pipeline {EXTRACTION_PIPELINE_EXTERNAL_ID} is empty")
    try:
        return Config.model_validate(yaml.safe_load(raw_config.config))
    except ValueError as e:
        logger.error(f"Invalid config: {e}")
        raise e



from __future__ import annotations

from pathlib import Path
from unittest.mock import MagicMock

import pytest
import yaml
from _pytest.monkeypatch import MonkeyPatch
from questionary import Choice

from cognite_toolkit._cdf_tk.commands.modules import ModulesCommand
from cognite_toolkit._cdf_tk.constants import BUILTIN_MODULES_PATH
from cognite_toolkit._cdf_tk.data_classes import Package, Packages
from tests.test_unit.utils import MockQuestionary


@pytest.fixture(scope="session")
def selected_packages() -> Packages:
    return Packages.load(BUILTIN_MODULES_PATH)


class TestModulesCommand:
    def test_modules_command(self, selected_packages: Packages, tmp_path: Path) -> None:
        assert selected_packages is not None

        target_path = tmp_path / "repo_root"

        cmd = ModulesCommand(print_warning=True, skip_tracking=True)
        cmd._create(organization_dir=target_path, selected_packages=selected_packages, environments=["dev"], mode=None)

        assert Path(target_path).exists()
        assert Path(target_path / "modules" / "infield" / "cdf_infield_common").exists()

    def test_modules_command_with_env(self, selected_packages: Packages, tmp_path: Path) -> None:
        assert selected_packages is not None

        target_path = tmp_path / "repo_root"

        cmd = ModulesCommand(print_warning=True, skip_tracking=True)
        cmd._create(
            organization_dir=target_path, selected_packages=selected_packages, environments=["dev", "prod"], mode=None
        )

        assert Path(target_path / "config.dev.yaml").exists()
        assert Path(target_path / "config.prod.yaml").exists()

    def test_config(self, selected_packages: Packages, tmp_path: Path) -> None:
        assert selected_packages is not None

        target_path = tmp_path / "repo_root"

        cmd = ModulesCommand(print_warning=True, skip_tracking=True)
        cmd._create(organization_dir=target_path, selected_packages=selected_packages, environments=["dev"], mode=None)

        config = yaml.safe_load(Path(target_path / "config.dev.yaml").read_text())
        assert config["variables"]["modules"]["infield"]["first_location"] == "oid"

    def test_adding(self, selected_packages: Packages, tmp_path: Path, monkeypatch: MonkeyPatch) -> None:
        target_path = tmp_path / "repo_root"
        cmd = ModulesCommand(print_warning=True, skip_tracking=True)

        first_batch = Packages({"infield": selected_packages["infield"]})
        second_batch = Packages({"quickstart": selected_packages["inrobot"]})

        cmd._create(organization_dir=target_path, selected_packages=first_batch, environments=["qa"], mode=None)
        with monkeypatch.context() as m:
            # Mocking questionary such that questionary.confirm.ask() returns True.
            questionary_mock = MagicMock()
            # MagicMock will always return other MagicMock objects
            # which when evaluated will return True.

            m.setattr("cognite_toolkit._cdf_tk.commands.modules.questionary", questionary_mock)
            cmd._create(
                organization_dir=target_path, selected_packages=second_batch, environments=["qa"], mode="update"
            )

        config = yaml.safe_load(Path(target_path / "config.qa.yaml").read_text())
        assert config["variables"]["modules"]["infield"]["first_location"] is not None
        assert (target_path / "modules" / "infield" / "cdf_infield_common").is_dir()

        assert config["variables"]["modules"]["inrobot"]["first_location"] is not None
        assert (target_path / "modules" / "inrobot" / "cdf_inrobot_common").is_dir()

    def test_add_without_config_yaml(self, tmp_path: Path, monkeypatch: MonkeyPatch) -> None:
        cmd = ModulesCommand(print_warning=True, skip_tracking=True)
        dummy_resource = "space: my_space"
        my_org = tmp_path / "my_org"
        moules = my_org / "modules"
        filepath = moules / "my_module" / "data_models" / "my.Space.yaml"
        filepath.parent.mkdir(parents=True, exist_ok=True)
        filepath.write_text(dummy_resource)

        def select_source_system(choices: list[Choice]) -> Package:
            selected_package = next((c for c in choices if "sourcesystem" in c.title.lower()), None)
            assert selected_package is not None
            return selected_package.value

        def select_sap_events(choices: list[Choice]) -> list:
            selected_module = next(
                (c for c in choices if "sap" in c.title.lower() and "event" in c.title.lower()), None
            )
            assert selected_module is not None
            return [selected_module.value]

        answers = [select_source_system, select_sap_events, False, False]

        with MockQuestionary(ModulesCommand.__module__, monkeypatch, answers):
            cmd.add(my_org)

        yaml_file_count = len(list(moules.rglob("*.yaml")))

        assert yaml_file_count > 1, "Expected new yaml files to b created"

        def select_sap_assets(choices: list[Choice]) -> list:
            selected_module = next(
                (c for c in choices if "sap" in c.title.lower() and "asset" in c.title.lower()), None
            )
            assert selected_module is not None
            return [selected_module.value]

        answers = [select_source_system, select_sap_assets, False, False]

        with MockQuestionary(ModulesCommand.__module__, monkeypatch, answers):
            cmd.add(my_org)

        new_yaml_file_count = len(list(moules.rglob("*.yaml")))

        assert new_yaml_file_count > yaml_file_count, "Expected new yaml files to be created"



import sys
from dataclasses import dataclass
from typing import Any

from cognite.client import CogniteClient
from cognite.client.data_classes._base import CogniteObject, CogniteResource, CogniteResourceList

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


@dataclass
class InstanceCounts(CogniteResource):
    nodes: int
    edges: int
    soft_deleted_nodes: int
    soft_deleted_edges: int

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            nodes=resource["nodes"],
            edges=resource["edges"],
            soft_deleted_nodes=resource["softDeletedNodes"],
            soft_deleted_edges=resource["softDeletedEdges"],
        )


@dataclass
class SpaceInstanceCounts(InstanceCounts):
    space: str

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            space=resource["space"],
            nodes=resource["nodes"],
            edges=resource["edges"],
            soft_deleted_nodes=resource["softDeletedNodes"],
            soft_deleted_edges=resource["softDeletedEdges"],
        )


class SpaceInstanceCountsList(CogniteResourceList):
    _RESOURCE = SpaceInstanceCounts


@dataclass
class CountLimitPair(CogniteObject):
    count: int
    limit: int

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            count=resource["count"],
            limit=resource["limit"],
        )


@dataclass
class InstanceCountsLimits(InstanceCounts):
    instances_limit: int
    soft_deleted_instances_limit: int

    @property
    def instances(self) -> int:
        return self.nodes + self.edges

    @property
    def soft_deleted_instances(self) -> int:
        return self.soft_deleted_nodes + self.soft_deleted_edges

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            nodes=resource["nodes"],
            edges=resource["edges"],
            soft_deleted_nodes=resource["softDeletedNodes"],
            soft_deleted_edges=resource["softDeletedEdges"],
            instances_limit=resource["instancesLimit"],
            soft_deleted_instances_limit=resource["softDeletedInstancesLimit"],
        )


@dataclass
class ProjectStatsAndLimits(CogniteResource):
    project: str
    spaces: CountLimitPair
    containers: CountLimitPair
    views: CountLimitPair
    data_models: CountLimitPair
    container_properties: CountLimitPair
    instances: InstanceCountsLimits
    concurrent_read_limit: int
    concurrent_write_limit: int
    concurrent_delete_limit: int

    @classmethod
    def _load(cls, data: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            project=data["project"],
            spaces=CountLimitPair._load(data["spaces"]),
            containers=CountLimitPair._load(data["containers"]),
            views=CountLimitPair._load(data["views"]),
            data_models=CountLimitPair._load(data["dataModels"]),
            container_properties=CountLimitPair._load(data["containerProperties"]),
            instances=InstanceCountsLimits._load(data["instances"]),
            concurrent_read_limit=data["concurrentReadLimit"],
            concurrent_write_limit=data["concurrentWriteLimit"],
            concurrent_delete_limit=data["concurrentDeleteLimit"],
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {
            "project": self.project,
            "spaces": self.spaces.dump(camel_case=camel_case),
            "containers": self.containers.dump(camel_case=camel_case),
            "views": self.views.dump(camel_case=camel_case),
            "dataModels" if camel_case else "data_models": self.data_models.dump(camel_case=camel_case),
            "containerProperties" if camel_case else "container_properties": self.container_properties.dump(
                camel_case=camel_case
            ),
            "instances": self.instances.dump(camel_case=camel_case),
            "concurrentReadLimit" if camel_case else "concurrent_read_limit": self.concurrent_read_limit,
            "concurrentWriteLimit" if camel_case else "concurrent_write_limit": self.concurrent_write_limit,
            "concurrentDeleteLimit" if camel_case else "concurrent_delete_limit": self.concurrent_delete_limit,
        }
